#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types

# Modules for plotting
import matplotlib
import matplotlib.pyplot as plt

'''sp_plot2008.py: generate a plot of a sapatt pattern in pseudo color

SYNTAX:   @@sp_plot2008,<pattern>,<ulevel>
EXAMPLES: @@sp_plot2008,zzz,0.04
'''

## Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def plotTheImage():
    '''plotTheImage does the work'''
    
    # ---- Constants ----
    '''
    Generally, the area (pixw x pixh) of the output image generated by
    matplotlib is meant to be presented on the screen of the computer
    on which the program runs. This means that pixelmaps are, in general,
    mapped down. There is a tip, which describes a way to keep the size
    of the pixel map.
    
    NOTICE The tipp does not preserve the pixel resolution in the plot,
           but at least there is some attenuation. I played with my_dpi
           and found the best resolution with 130.
    '''
    my_dpi = 106.8

    logscale = False # May become a parameter
    
    argus = len(sys.argv) - 1
    if argus < 1:
        return "Script needs arguments"
    
    arglist = argexpand( 3 ) # expand to 3 arguments

    ## (1) The image file
    thefile = arglist[0]
    b = sp.Patt()
    if b.load( thefile ):
        return 'Loading error'
    draw = b.copy()
    
    if argus < 2:
        ulevel = np.max(b.mapp)
        print('Upper level from image: %10.3e'%ulevel)
        if ulevel <= 0:
            return 'Upper level must be positive'
    else:
        try:
            ulevel = float(arglist[1])
        except:
            return 'Could not convert "%s" to float'%arglist[1]
    
    if logscale:
        draw = draw._scale_log_pos( ulevel )

    # Apply modulo function to the mapp
    draw.mapp = np.fmod(draw.mapp,ulevel)

    # The wrapper for a complex output image: a scattering pattern ("sp")
    # defines the aspect ratio of the resulting figure
    w, h = draw.width, draw.height
    # The overall figure
    # fig = plt.figure(figsize=(w/my_dpi, h/my_dpi), dpi=my_dpi)
    fig = plt.figure(figsize=(6, 6), dpi=my_dpi)
    # If we want to decorate the figure, we must not plot directly into it,
    # but define a set of subplots which contains only one single subplot 
    # arranged in a 1x1 matrix and labeled subplot 1.
    ax1 = fig.add_subplot(111)
    # Because we have 1x1 subplot(s), the definition is already complete.
    
    # We fill in the scattering pattern
    ax1.imshow( draw.mapp,vmin=0,vmax=ulevel,cmap='CMRmap',\
        interpolation='none' )
    
    plt.axis('off')
    ax1.axes.get_xaxis().set_visible(False)
    ax1.axes.get_yaxis().set_visible(False)
    # NOTICE bbox_inches removes rest of the white space around.
    # Newer versions of matplotlib require =0 instead of ='tight'
        
    # Plot the figure
    # fig.tight_layout(pad=0) NOTICE impossible for image
    # plt.savefig('plot.png')
    # Make file name for output
    fnbody = thefile.split(".")
    fnpng  = fnbody[0]+'.png'
    print('Writing picture to "%s"'%fnpng)
    plt.savefig( fnpng, bbox_inches='tight', pad_inches=0, dpi=my_dpi )
    plt.show()
    plt.close(fig) # free memory

    return ''

#END plotTheImage()

# END script procedures

# Main:
err = plotTheImage()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()



