#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types
from sapiawin import *

'''sp_prep2008.py: Detector patterns are already converted by cbf2sapatt!

   Next script to call is:       sp_harm2008.py,<frame>

NOTICE All preprocessing is in detector units. Conversion to units
of the scattering vector is done in ??? WARNING sp_corr2harm.py

Gültigkeitsbereich: Datenvorverarbeitung der Experimente
zur Relaxation nach schlagartigem Strecken bzw. Entspannen
von TPU-Elastomeren, die im August 2020 an der P03 am DESY
durchgeführt wurden. Das waren Produktionsexperimente, aber
noch mit einem Test-Aufbau. Deshalb ist dieses Programm
ein Einmal-Programm, dass nur für diese eine Messzeit relevant
ist.

Experimentplan: Die Proben werden mehrmals geschlagen. Um nicht
zu viele Daten aufzunehmen, werden nur 1500 SAXS-Bilder auf-
genommen (1000 Bilder im ersten Dehn-Relax Zyklus und etwas
"Fleisch" drumherum). Das Schlagen geht weiter und kurz vor dem
5. Schlag startet dasselbe Programm noch einmal, um die Reaktion
auf den 5. Schlag zu beobachten.

Drei Prozesse wurden parallel gefahren. Jeder Prozess erzeugt
seine eigenen Messdaten. Die Prozesse sind nicht intrinsisch
gekoppelt.

Detektorprozess: Der Detektorprozess erzeugt SAXS-Bilder im 
cbf-Format. Die Dateien enthalten keine Umgebungsdaten (nicht
einmal den Synchrotronstrom). Die Zykluszeit von 20 ms wird
sehr gut eingehalten. Zum Handling einer Serie müssen detmask und
Maschinenuntergrund bereit gestellt werden. Das erfolgt durch
Programmierung, nicht durch Einlesen (Einmal-Programm).

Monitorprozess: Der Monitor ist ein "DUMMY-Scan" der Mess-
streckensteuerung. Er läuft nebenbei. Er loggt den Synchrotron-
strom, die Ionisationskammer und die PIN-Diode im Beamstop indem
er die VFC-Zähler (Zähler hinter den 8-Kanal Voltage-to-Frequency-
Convertern) jeweils 1 s laufen lässt, dann ausliest und in eine
Zeile einer fio-Datei schreibt. Das Auslesen und Schreiben braucht
ca. 300 ms. Der Prozess mittelt also über 50 Streubilder, lässt
danach 15 Bilder aus und mittelt über die nächsten 50 Streubilder.
Die Daten zeigen, dass sich die PETRA-Intensität während eines
Relaxationszeitraums deutlich ändert, aber über 1 s recht stabil
ist. Eine Normierung der SAXS-Intensität müsste mit individuellen
Schmiegekurven gemacht werden, um Artefakte durch Sprünge zu
vermeiden. Da tolerieren wir lieber Artefakte durch I0-Variation.
Die Absorptionsfaktoren aber lassen sich (nach trickreicher Analyse
der fio-Dateien) sehr gut bestimmen. Dazu dient @@beatfio.py
Die Prozedur muss vorher laufen und schreibt die relevanten Daten
in Dateien der Art <serie>.beat

Schlagprozess: Der Schlagprozess erzeugt Dateien "Imp_hhmmss.txt".
Die Zeit ist die Startzeit des Schlagprozesses. Der Schlagprozess
startet unabhängig vom Detektorprozess. Um das Uhrenproblem in den
Griff zu bekommen, wird die Schlagdatei am Anfang so gekürzt, dass
die erste Zeile dem Nullzeitpunkt des 1up-Schlages entspricht.
Die gekürzte Datei wird als Imp_hhmmss.dat abgespeichert und
hier von diesem Programm gelesen. Die erste Zeile ist dann
Referenzzeit und 10 s später folgt der 1dn-Schlag.
Neben der Uhrzeit ist einzig die Kraft (in Spalte 3) wichtig.
Darin liegt der Wert der Imp-Dateien.

Der Detektor ist ein PILATUS 300K, der sehr gut positioniert ist
und erstaunlich viel vom SAXS-Bild erfasst. Abstand: 2700 mm.
Wellenlänge: 0.1 nm

WAXS wurde in separaten, mechanisch gleichartigen Experimenten gemessen.

Carry out a background subtraction by
    (1) dividing each pattern by its exposure ( geta('exposed') )
    (2) compensating the absorbed intensity using emut
    (3) subtracting the machine background
    (4) masking by the detector mask
    (5) storing the result in a file with its name generated
        from <rellabel>+_####.sjs

SYNTAX:   @@sp_prep2008,<impfile>,<detzero>,<beatframe>,<rellabel>
EXAMPLE:  @@sp_prep2008,Imp_153952,88,csa17e_1_00088,qa17e_s1up
          @@sp_prep2008,Imp_153952,88,csa17e_1_00589,qa17e_s1dn

    <impfile>:    Die vorn gekürzte Text-Datei, in die die Kraftwerte
                  gedumpt wurden. Der Dateiname genügt der Konvention:
                  Imp_hhmmss.dat. Durch die Kürzung kann dieses Programm
                  den Versatz zwischen Detektorzeit und Schlagzeit
                  herausrechnen. Für die Testdatei sind das 827 ms!!!
    <detzero>:    Detektor-Frame des 1up-Schlags (für den Uhrenvergleich)
    <beatframe>:  Das Detektorbild, in dem der Schlag sichtbar wird
                  (an dem die nun folgende Relaxation von max.
                  500 Frames hängt)
    <rellabel>:   Labelstring, der Material und Schlagtyp charakteri-
                  siert. Z.B. "_1up" oder "_5dn" - letzteres für die
                  Relaxation im erschlafften Zustand nach dem 5. Schlag.
                  Das Label wird nicht nur zum Erzeugen der Dateinamen
                  verwendet sondern auch, um die passenden Umgebungsdaten
                  zuzuordnen.
                  
                  Nomenklatur: q: campaign letter
                               a17e: annealed, sample 17, run e
                               _s1up: SAXS, relaxation after 1st up-beat

Für diese Messzeit hart kodiert sind:
    <bgfile>:    NOTICE bg_saxsm3.sjs (Der SAXS-Untergrund median3-geglättet)
    <maskfile>:  NOTICE detmask3.sjs (Die SAXS-Maske, 3 Pixel erodiert)
    Weitere benötigte Dateien:
    NOTICE csa17e_00001.beat - Die aus der .fio-Datei mittels
           @@beatfio,csa17e_00001 erzeugte Datei mit den Daten der
           Strahlmonitore (in /fioSAXS/)
    NOTICE Imp_hhmmss.dat - Aus der Originaldatei Imp_hhmmss.txt erzeugte
           Schlagdatendatei. Vorn sind die Kopfzeilen und alle Zeilen bis
           zum Beginn des 1up-Schlags gelöscht. Die Uhrzeit der 1. Zeile
           ist dann die Uhrzeit, die die Schlaguhr zum zum Zeitpunkt zeigt,
           an der die Detektoruhr das erste Streckbild markiert hat.
           
    
    Lage: AnaSAXS/Leer bzw. AnaSAXS/Maske (dort auch csa17e_1_#####.sjs)
'''

## Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

    #pia = piawins( self.piawinds, verbose=True )
    self.piawin = piawins( self.piawinds )

def preprocess():
    '''preprocess does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    ### Fetch hard-coded data
    ##  Detektor mask
    maskfile = 'detmask3.sjs'
    mask = sp.Patt()
    err = mask.load( maskfile )
    if err:
        return '%s: reading failed'%maskfile    

    ##  Machine background
    empfile = 'bg_saxsm3.sjs'
    emp = sp.Patt()
    err = emp.load( empfile )
    if err:
        return '%s: reading failed'%empfile
    '''
    # Die Datei ist schon auf Belichtung 1 s normiert
    expo = emp.geta('exposed')
    emp.mapp = emp.mapp / expo
    emp.logentry('Intensity normalized to unit exposure')
    emp.puta('exposed',1)
    '''
    # bg_saxs2_table.ods:
    i0emp = 77522.0 # I0-Counter. Mittelwert bezogen auf 1 s Belichtung. 
    i1emp = 177037.0 #PIN-Diode Counter. Mittelwert/s aus 10 Werten Langzeitbelichtung
    
    # Zykluszeit Detektor
    detector_cycle = 0.020 # ... in Sekunden
    
    ### End: Fetch hard-coded data
    ### --------------------------------------------
    
    ### Argument parsing
    argus = len(sys.argv) - 1
    if argus != 4:
        return "Script needs 4 args (impfile,detzero,beatframe,rellabel)!"
    
    ## (2) detzero
    dzs = sys.argv[2]
    try:
        detzero = abs(int(dzs))
    except:
        return 'detzero "%s" is no integer'%dzs

    ## (1) impfile vorhanden?
    impfn = sys.argv[1]
    if len(impfn.split('.')) == 1:
        impfn = impfn + '.dat'
    try:
         impf = open(impfn, "r")
    except:
        return '%s kann nicht geöffnet werden'%impfn
    # Schlaguhr setzen
    line = impf.readline()
    words = line.split()
    hms = words[0].split(':')
    #NOTICE Stand der Schlaguhr beim ersten Schlag.
    schlagnull = (int(hms[0])*60.0+int(hms[1]))*60+float(hms[2])
    impf.close()
        
    ## (4) Nomenklatur des Relaxationslabels
    rellabel = sys.argv[4]
    isup = rellabel[-2:] == 'up'
    isdn = rellabel[-2:] == 'dn'
    # Am Schluss von rellabel muss entweder "up" oder "dn stehen. Dieser
    # Designator zeigt an, ob die Probe gerade gedehnt ("up") oder 
    # erschlafft ("dn") worden ist.
    if not (isdn or isup):
        return 'rellabel does not end with "up" or "dn"'
    
    ## (3) Existenz des Schlagframes
    beatframe = sys.argv[3]
    frp = sp.Patt()
    err = frp.load( beatframe )
    if err:
        return 'beatframe: File reading failed'
    ## Die Serie: alles hinter letztem Unterstrich abgetrennt
    ser = '_'.join(beatframe.split("_")[:-1])
    ## Index des ersten Eintrages extrahieren (der ist nicht immer 00000)
    sfif = serfiles[0].split('_')[-1:][0] # noch mit Dateiextension
    sfif = int(sfif.split('.')[0])

    ## Einschub: Bestimmung Erstschlagzeitbestimmung
    ## (Detektor-Uhrzeit des Erstschlags bei detzero
    zeroframe = ser+'_0'+str(10000+detzero)[1:]
    err = frp.load( zeroframe )
    if err:
        return 'detzero frame "%s": File reading failed'%zeroframe
    # detnull wird die second-of-day des allerersten Schlags (zum vgl. mit schlagnull)
    rec = frp.geta('recorded')
    detnull = (rec[3]*60.0+rec[4])*60+rec[5]+rec[6]*0.001    
    ## Ende Erstschlagzeitbestimmung

    ## Liste der Streubilder der ganzen Serie zu der beatframe gehört
    serfiles = sorted(glob.glob(ser+"_*.sjs"))
    # Wenn beatframe existiert sollte auch die ganze Serie existieren
    # beatframe numerisch
    bfs = beatframe.split('_')[-1:][0] # String hinter letztem Querstrich
    # Hängt da noch eine Datei-Extension hinten dran?
    test = bfs.split('.')
    if len(test) > 1:
        bfs = test[0]
    bfnum = int(bfs)
    # WARNING  Die nächste Zeile setzt voraus, dass die Frames beginnend
    #          mit dem Ettikett sfif in 1er-Schritten (ohne Lücke)
    #          vorliegen. Ohne "-sfif" müssen sie sogar mit _00000 anfangen.
    relaxf = serfiles[bfnum-sfif:] # Relaxation vom Schlag bis zum Serienende
    # NOTICE Spätestens nach 501 Frames kommt der nächste Schlag
    if len( relaxf ) > 500:
        relaxf = relaxf[:500]
        
    #NOTICE Option: Abbruch bei sehr unvollständigen Serien
    #if len( relaxf ) < 100:
    #    return "Less than 100 pattern files in relaxation monitor"
    
    # WARNING NOTICE TODO Nur zum Testen sollen nur 3 Bilder
    #                     bearbeitet werden
    #relaxf = relaxf[:3]
    
    # relaxf enthält jetzt alle Dateien zur Relaxation nach dem Schlag
    # und die Relaxationszeit steckt in der Sequenznummer (Schrittweite: 20 ms).
    # Das erste relaxf ist das beatframe. Dazu gehört die Relaxationszeit 0.
    
    # Bevor es in die Schleife geht müssen noch die beiden Absorptionslevel
    # berechnet werden. Notwendige Werte sind in den beat-Dateien (@@beatfio,...).
    fiobeat = beatframe.split('_')[0]+'_00001.beat' # NOTICE Konvention
    try:
        beatf = open( fiobeat )
    except:
        return 'Cannot open %s'%fiobeat
    ionis = beatf.readline().split() # Zählerstände 
    times = beatf.readline().split()
    beatf.close()
    
    i0samp = float(ionis[0]) # Counts in Ioni0 nach 1 s Belichtung mit Probe
    i1samp_dn = float(ionis[1]) # Bei schlaffer Probe: Counts in PIN-Diode
    i1samp_up = float(ionis[2]) # Bei straffer Probe: Counts in PIN-Diode
    # Berechnung der Absorptionsfaktoren
    emut_dn = i1samp_dn*i0emp/(i1emp*i0samp)
    emut_up = i1samp_up*i0emp/(i1emp*i0samp)
    # Auswahl des richtigen Absorptionsfaktors
    emut = emut_dn
    if isup:
        emut = emut_up
    # factor proportional to sample thickness
    tfac = -5.7*np.log(emut)
    tmul = 1/tfac # thickness change multiplicative compensation factor
        
    # TODO Normierung auf schwankende Synchrotronintensität wäre grundsätzlich
    #      hilfreich. Leider sind die Daten in der fio-Datei nur grob gerastert.
    # Alle 1300 ms gibt es einen Eintrag in der fio-Datei. Dieser Eintrag
    # gehört also einer Gruppe von 65 Streubildern. An der Gruppengrenze treten
    # deshalb wohl störende Stufen auf, die die Korrektur hintertreiben. Auch
    # ist nicht klar, ob die fio-Uhr und die Detektor-Uhr dieselbe Zeit anzeigen.
    # Wir wissen, dass es zwischen Det-Uhr und Schlag-Uhr eine Differenz in der
    # Größenordnung von 1 s gibt. Auch ist im fio-Bild nur die Startzeit des
    # ganzen DUMMY-Scans einmal dokumentiert (ohne Sekundenbruchteile).
    # Weiterzählen geht nur ungefähr in Schritten von 1300 ms.
    # fiosod = float(times[0]) # ist von @@beatfio in Einheit sod: "Tagessekunde"
    
    # Voreinstellungen für die Arbeitsschleife
    # 
    # Versatz zwischen Detektoruhr und Schlaguhr
    det2imp = schlagnull - detnull # imp_sod = det_sod + det2imp
    
    # Über alle Frames des Relaxationszweiges
    for i, thefile in enumerate( relaxf ):
        reltime = detector_cycle*i # Relative Zeit im Relaxationszweig in s
        if frp.load( thefile ):
            return '%s: File reading failed'%thefile

        frp.logentry('@@sp_prep2008.py --- Start pre-evaluation')
        frp.logentry('... from raw data %s'%thefile)
        expo = frp.geta('exposed')
        frp.mapp = frp.mapp / expo
        frp.logentry('Intensity normalized to unit exposure')
        frp.puta('exposed',1)
        frp.mapp = frp.mapp / emut
        frp.logentry('Absorption elimination: Dividing the intensity by emut = %5.3f'%emut)
            
        # Subtract machine background
        frp.mapp = frp.mapp - emp.mapp
        frp.logentry( 'Machine background "%s" subtracted'%empfile )
        # Correct for unit sample thickness
        frp.mapp = frp.mapp * tmul
        frp.logentry( 'Intensity multiplied by %6.3f for thickness compensation '%tmul )

        frp.logentry( 'Cosmic ray outliers removed by median 3x3' )
        frp.median(3)
        
        # Check identity of matrix dimensions
        if frp.mapp.shape != mask.mapp.shape:
            return 'Different dimensions of scattering pattern and detector mask'
        # Mask invalid pixels
        frp.mapp = frp.mapp * mask.mapp
        frp.logentry( 'Masked by detector mask "%s"'%maskfile )

        # Relaxationszeit eintragen
        frp.puta('relaxtime',reltime)
        frp.logentry( 'Attribute relaxtime set to %6.3f seconds'%reltime )
        
        # Kraft zum Bild eintragen (leider sind die Kraftwerte nur sporadisch
        # gelistet, so dass zu jedem Bild nur maximal 2 passende Kraftwerte
        # da sind)
        # NOTICE second-of-day der Speicherung des Detektor-Bildes berechnen
        rec = frp.geta('recorded')
        det_sod = (rec[3]*60.0+rec[4])*60+rec[5]+rec[6]*0.001
        det_sod = det_sod + det2imp #NOTICE abgebildet auf die Schlaguhr
        
        # TODO Anna sagte, dass die Kraftwerte manchmal nicht stimmen.
        #      Schau' in ihr Spreadsheet!!! (8 N falsch ...)

        # Jetzt aus der Imp_hhmmss.dat einen passenden Kraftwert picken
        impf = open(impfn, "r") # Vorhandensein ist schon getestet
        olddelta = 20.0 # Riesig: "20 s zu früh"
        for line in impf:
            words = line.split()
            hms = words[0].split(':')
            search_sod = (int(hms[0])*60.0+int(hms[1]))*60+float(hms[2])
            if  search_sod >= det_sod:
                break # Kraftwert gefunden
            # Alten Wert speichern -- vielleicht ist er näher 'dran
            olddelta = det_sod - search_sod
            oldwords = words
        impf.close()
        delta = abs(search_sod - det_sod) # abs brauchen wir eigentlich nicht
        # Trag' den Kraftwert ein, der näher an der
        # Detektorzeit liegt
        if delta > olddelta:
            words = oldwords # war näher dran
        force = float(words[2])

        print( "Force found: %s %f"%(words[0],force) )
        # frp.puta('force_clock', words[0])
        frp.logentry('Considering clock difference [imp - detector] = %6.3f s'%det2imp)
        frp.puta('force', force )
        frp.logentry('Force from %s. Imp-clock: %s is %7.3f N'%(impfn,words[0],force))
        
        # Bookkeeping

        # Neuen Dateinamen bilden
        framestring = str(10000+i)[1:] # 4stellig mit führenden Nullen
        newfn = rellabel+'_'+framestring
        frp.puta('fn',newfn)
        frp.logentry('New name for pre-evaluated framefile: %s'%newfn)
        frp.logentry('@@sp_prep2008.py --- End pre-evaluation')
        frp.save( newfn )
    # Ende Arbeitsschleife
    return ''
#END preprocess()

# END script procedures

# Main:
err = preprocess()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()



