#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types

'''wf_map_sk.py: Mapping of a WAXS fiber pattern from the detector
                 plane (p1,p3) to the representative plane (s12,s3)
of reciprocal space. Although this version requires to be run under
saskia, there are not many references to saskia. All the functional
methods are defined here in order to ease implementation into a
different python environment.
              
How to call the program from saskia:

SYNTAX:   @@wf_map_sk,<saveset>,<framefile>[.sjs],<beta>
EXAMPLE:  SASkia> @@wf_map_sk,zzsav,hart_qm23_0000

Result: The mapped WAXS or WAXD pattern is written a file of flavor "hars", in
        the example: hars_qm23_0000.sjs

<saveset>:   The saveset is a text file. In its first line it must contain 3
             or 4 numbers, separated by blanks:
                <wavel>  : The wave length of the X-rays (in Å or nm)
                <dspacr> : The d-spacing (in the same units as <wavel>) of
                           of the following <radius> of a reflection circle
                           on the detector
                <radius> : radius of a reflection circle in pixel units
                <beta>   : fiber tilt angle in degrees. This value will only
                           be used, if it is not overwritten by supplying it
                           as the third argument of the call.
                           WARNING It may be necessary to replace beta = -beta
                           in the program. The tilt-angle correction has not 
                been tested in the Python-environment but simply copied from
                the old PV-WAVE code.
<framefile>: The file which contains the WAXS pattern which shall be mapped
             on the (s12,s3)-plane. It must be stored in a quadratic matrix
             in (p1,p3), but blind pixels (intensity = 0) are allowed. 
             the edge length of the matrix (p_len) must be an even number.
             The center of the pattern must be in the center of the matrix
             (in the crossing point between the 4 central pixels). The
             pattern must be aligned with the fiber axis p3 vertical. The
             pattern must show mirror symmetry about the fiber axis, but
             if the fiber is inclined no more symmetry is required. The pixels
             themselves must be quadratic. Unimportant is the edge length
             of a pixel on the detector.
             The FILENAME must conform to a special naming convention:
             <flavor>_<experiment_and_frame>.sjs -- this is, because the output
             file gets a new flavor, namely "hars" (harmonized pattern in "s").
RESTRICTIONS:
    (1) It may be more convenient to replace beta = -beta in the code.
    (2) The polarization factor is not used assuming unpolarized X-rays.
        If the polarization quality (qp) is known and polarization effects
        are seen in the pattern, then qp must be set correctly and the
        respective line of code must be activated. The line is close to the
        end of this file.
PROCEDURE:
    Mapping works only on the right side of the pattern in reciprocal space.
    At the end the left side is generated from mirror symmetry.
    
    (1) The maximum area that can be mapped into s-space is given by the
        maximum (and minimum) reflection circles on the detector. A corresponding
        matrix in s-space is set up as a mask. We set it 1 ("mapable") everywhere.
        From the mapables we remove by setting pixels to 0:
        - s-values outside the outer reflection circle
        - s-values inside an inner reflection circle (if there is one)
        - s-values in the wedges on the meridian (my technique by killing 
          circular disks erases the left side of (the s-matrix completely)
    (2) The remnant mapables are collected in a collection of s-values "svaltu"
        which, in fact, is a Python-tuple-construct. For each of the mapables its
        corresponding partner on the detector image is determined (this is a
        pointer from s-plane into detector plane). p1map(svaltu), p3map(svaltu)
        describe the mapping.
    (3) Collection cleanup: Some of the pointers point to positions which are
        outside the detector-pattern matrix. They are removed.
    (4) Intensities of the mappable pixels are copied into a pixel matrix
        preset by 0 ("blind pixel")
    (5) Intensities are corrected using the Jacobian of the reverse mapping
    (6) Intensities are corrected for the angular variation of the absorption
    (7) This was the right half. The left half is generated by transposing,
        rotating and adding.
NOTICE  In materials science we often have sets of frames with the same mapping.
        Then the preparation is independent of the individual frame and the same
        svaltu, p1map, p3map, Jacobian, ... can be used for all frames speeding
        up the process considerably.
'''

## BEGIN Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
# END argexpand()

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
# END banner()

def Distquad( n ):
    '''The result is a quadratic matrix containing the distance of the pixel
    from the upper left corner of the matrix assuming that the coordinate
    origin is in the upper left corner of the upper left pixel (leading to the
    common definition in the field of scattering).
        This function is a relative of the FUNCTION DIST() of IDL and PV-WAVE.
    There the origin is in the CENTER of the corner pixel and, finally, 
    4 quadrants are arranged in such a way that in the complete matrix each
    quadrant measures the distance of each pixel from its nearest CORNER.
        In scattering one needs the distance of each pixel from the CENTER of
    a scattering or diffraction pattern which has its origin in the center of
    the complete matrix. There the center is in the crossing point of the 
    4 quadrants, one of which is the result of Distquad().
    
    EXAMPLE: Distquad(4):
    [[0.70710678 1.58113883 2.54950976 3.53553391]
    [1.58113883 2.12132034 2.91547595 3.80788655]
    [2.54950976 2.91547595 3.53553391 4.30116263]
    [3.53553391 3.80788655 4.30116263 4.94974747]]
    
    WARNING Remember the "Chinese" addressing of matrices in Python:
    matrix[y,x] with y running downwards and x running from left to
    right.
    '''
    # axis = np.linspace( 0, n-1, n ) # standard IDL or PV-WAVE convention
    axis = np.linspace( 0.5, n-0.5, n ) # Scattering convention 
    return np.sqrt(axis**2 + axis[:,np.newaxis]**2)
# END Distquad()

def Patdist( n ):
    ''' n is an even and positive integer. patdist(n) returns a matrix in
    which each element contains the Euclidian distance of each (quadratic)
    pixel from its center. By convention the center is in the crossing point
    of the four quadrants.
    Patdist(4):
    [[2.12132034 1.58113883 1.58113883 2.12132034]
     [1.58113883 0.70710678 0.70710678 1.58113883]
     [1.58113883 0.70710678 0.70710678 1.58113883]
     [2.12132034 1.58113883 1.58113883 2.12132034]]    
    '''
    # n2 is n/2
    n2 = np.abs(int(n+0.5))/2
    n = 2*n2 # NOTICE enforce positive and even integer
    
    # make the result matrix
    euklid = np.empty((n,n))
    # compute a quadrant of the result
    euquad = Distquad(n2)
    # Because of Python's Chinese addressing, with y being the
    # first coordinate running downwards, this is the lower
    # right (lr) quadrant
    euklid[n2:,n2:] = euquad # lr lower right
    # fill the other quadrants by properly rotated copies
    euklid[:n2,n2:] = np.rot90(euquad) # ur
    euklid[:n2,:n2] = np.rot90(euquad,2) # ul
    euklid[n2:,:n2] = np.rot90(euquad,3) # ll
    return euklid
# END Patdist()

def killdisk( img, x, y, r ):
    '''
    Erase the intensity on a circular disk in the scattering pattern.
    Pattern and disk may overlap even only partially. In my PV-WAVE
    programs this script was sf__killcircle

    x, y :  Position of the center of the circle in pixel units. The
            lower left pixel is considered (0,0). This means 
            (1) that the information on the center of a scattering pattern
                must have been considered by the caller before, but
            (2) that the Chinese addressing of matrices is considered here.
    r:  radius of the circle in pixel units
    '''
    dia2 = 2*r # diameter of required mask
    
    # claimed disk area in img
    xu = x - r
    xo = x + r
    yu = img.height - y - r # running Chinese
    yo = img.height - y + r # running Chinese
    
    # Generate a distance mask of the disk
    disk = Patdist( dia2 )
    # Make it a mask with 0 (black) inside and 1 (white) outside
    disk = disk > r
    
    # Compute the intersection of disk with the image
    # so that xu,yu, xo, yo define the position of a
    # (truncated) patch which liese within img
    upx = img.width - 1
    upy = img.height - 1
    # adjust lower boundary
    if xu < 0:
        # WARNING Remember Chinese addressing
        disk = disk[:,-xu:]
        xu = 0
    if yu < 0:
        # WARNING 0 is top (untouched), upy+yu is new bottom
        disk = disk[-yu:,:]
        yu = 0
    if xo > upx:
        disk = disk[:,:upx-xu]
        xo = upx
    if yo > upy:
        # WARNING Chinese: subtract y0-upy from upper limit
        disk = disk[:-yo+upy,:]
        yo = upy

    # Does the disk portrude into the scattering pattern?
    # print(disk.shape) # For testing
    # print('(%i .. %i, %i .. %i)'%(xu,xo,yu,yo))
    if (disk.shape[0] > 0) and (disk.shape[1] > 0):
        # ... then erase the intensities under the disk
        img.mapp[yu:yo,xu:xo] = disk * img.mapp[yu:yo,xu:xo]
    return img
# END killdisk()

def killoutdisk( img, x, y, r ):
    '''
    Erase the intensity OUTSIDE a circular disk in the scattering
    pattern. Pattern and disk may overlap even only partially.
    In my PV-WAVE programs this was sf__killoutcircle

    x, y :  Position of the center of the circle in pixel units.
            The lower left pixel is considered (0,0). This means 
            (1) that the information on the center of a scattering pattern
                must have been considered by the caller before, but 
            (2) that the Chinese addressing of matrices is considered here.
    r:  radius of the circle in pixel units

    '''
    # Build a mask of suitable size, which is one everywhere
    mask = img
    mask.mapp = 0.0*img.mapp + 1.0
    # In the mask: Kill everything INSIDE the requested disk
    killdisk( img, x, y, r )
    # Invert the mask
    mask.mapp = 1.0 - mask.mapp
    # .. and compute the required result
    img.mapp = img.mapp * mask.mapp
    return img
# END killoutdisk()

def _rctos( wavel, samp, radius ):
    '''Convert the radius of the reflection circle on the detector
    to the radius of the reflection circle in reciprocal space
    (s-space).
    samp  : distance from sample to the "burn spot" on the detector.
    radius: radius on detector (in the same length units as samp)
    wavel : wavelength of radiation (in reciprocal units of s-space)
    '''
    theta = 0.5*np.arctan2( radius, samp )
    svalue = 2*np.sin(theta)/wavel
    return svalue
# END _rctos()

def map_it():
    ## BEGIN Retrieve the arguments
    
    # Number of arguments supplied by the call
    NumArgs = len( sys.argv ) - 1
    arglist = argexpand( 3 ) # Expand for 3 arguments

    ## (1) The save set 
    thesaveset = arglist[0]
    try:
        ssf = open( thesaveset )
    except:
        return 'Cannot open %s'%thesaveset
    saved = ssf.readline().split() # list of space separated strings
    ssf.close()
    # Read saveset variables
    wavel  = float(saved[0])
    dspacr = float(saved[1])
    radius = float(saved[2]) # WARNING was number 22 in PV-WAVE
    
    ## (2) The image file
    thefile = arglist[1]
    img = sp.Patt()
    if img.load( thefile ):
        return 'Loading error'
    if img.width != img.height:
        return 'Pattern must be quadratic'
    if not img.hasa('center'):
        # ... and centered
        return 'Center attribute missing'
    # center of the CENTERED, QUADRATIC pattern
    icenx, iceny = img.geta('center')
    if (icenx != iceny) or (icenx*2 != img.width):
        return 'Center is not in the middle of the pattern map'
    p_cen = icenx   # Center of the quadratic pattern in pixels
    p_len = 2*icenx # Edge length of the quadratic pattern in pixels
    
    '''
    NOTICE WAXS patterns to be mapped on the fiber plane sometimes
    show harmony (4 identical quadrants). During this processing the
    harmony is temporarily lost. If this is not indicated by deleting
    the corresponding attribute, the pattern will auto-harmonize when
    its intermediate state is saved (by fibr.save('dummy')). So:
        delete the attribute if it is present!
    '''
    # NOTICE mapping destroys harmony permanently, if beta != 0
    if img.hasa('Compressible1/4'):
        img.dela('Compressible1/4')

    ## (3) Tilt angle beta
    if NumArgs < 3:
        if len( saved ) > 3:
            betadeg = float(saved[3]) # WARNING was number 24 in PV-WAVE
        else:
            return 'beta not found in save set'
    else:
        try:
            betadeg = float(arglist[2])
        except:
            return 'Tilt angle beta: Could not convert %s to float'%thesaveset
    beta = np.pi*betadeg/180

    # Log input data to user screen
    print( '%6.3f [Å or nm] X-ray wavelength'%wavel )
    print( '%6.3f [Å or nm] picked d-spacing'%dspacr )
    print( '%6.1f [pixels]  radius corresponding to d-spacing'%radius )
    print( '%6.2f [°]       fiber tilt angle beta'%betadeg )
    
    #**************************************************************
    # Compute the size of the output map. The size shall always be
    # related to the WIDTH of the input image, not to the actual
    # actual maximum s-value in the input data 
    #**************************************************************

    sr = 1.0/dspacr # Radius of the supplied reflection ring in s-space
    thetar = np.arcsin(0.5*wavel/dspacr) # ... its half scattering angle

    # Distance between sample and detector in pixel units
    samp = radius/np.arctan(2*thetar)
    # print('%i [pixels] distance between sample and detector'%samp)
    # Upper limit of s1 (in horizontal, i.e. equatorial direction
    # which can be mapped into an image with its width in pixels
    # being the same as that of the input image 
    s1max = _rctos(wavel,samp,p_cen)
    print( '%6.3f [1/Å or 1/nm] maximum s12 in image to be mapped'%s1max )

    # Generate the result image
    fibr = img.copy()
    # Define the grid of the output map fibr in reciprocal space
    del_s = 2*s1max/(fibr.width-1)
    # Bookkeeping for sapatt-patterns:
    #     vaxras: Vector of AXes del_s (x, y, f[y,x])
    fibr.vaxras[0] = del_s # for s12-axis
    fibr.vaxras[1] = del_s # for s3-axis

    # ***************************************************************
    # * Determine blind areas for the process and prepare some
    # * matrices ("maps") -- all to speed up mapping
    # ***************************************************************
    # valid intensity pixels in input pattern
    valii = np.where( img.mapp > 0 )
    # Generate the map of the Euclidian Distance of each pixel
    # from the center if the image in units of pixels
    MEukl = Patdist(p_len)
    # Minimum and maximum "reflection" radii based on the innermost
    # and outermost pixels containing valid intensity on the detector
    # in pixel units. 
    minrad = int(np.min(MEukl[valii]) + 0.5) # round into valid region
    maxrad = int(np.max(MEukl[valii]) - 0.5) # round into valid region
    print( 'Valid pixels between radii %i and %i'%(minrad,maxrad) )
    # TODO del_s anpassen (dann auch fibr.vaxras[]!)???
    # Map these radii into radii in reciprocal space
    minsrad = _rctos(wavel,samp,minrad)
    maxsrad = _rctos(wavel,samp,maxrad)
    print( 'Valid s-values between %6.3f and %6.3f'%(minsrad,maxsrad) )

    # matrix of the magnitudes of the scattering vector in (s1,s3) plane
    del_s = np.double(del_s)
    Mslen = del_s * MEukl

    ## Make a pick-mask that has values only where there is 
    # a chance to compute s-space intensities from measured
    # data
    smask = np.ones((p_len, p_len),dtype=int)
    ## (1) Keep only the possible range of reflection rings
    smask[ np.where(Mslen <= minsrad) ] = 0
    smask[ np.where(Mslen >= maxsrad) ] = 0
    ## (2) Consider the blank areas due to fiber tilt
    ##     They are cut-out by two circles
    # Compute the s3-value corresponding to 2*beta - This
    # is the point where the two circles intersect in the pattern
    s3beta = 2*np.sin(beta)/wavel
    # Coordinates of the centers of the intersecting circles
    s3c = s3beta/2
    s1c = np.sqrt(1/(wavel*wavel) - s3c*s3c) # +s1c and -s1c
    # turn s-values of center coordinates into pixels
    p3c = int(s3c/del_s) + p_cen
    p1c = int(s1c/del_s) + p_cen
    print('(±p1c,p3c) = (±%i,%i)'%(p1c,p3c))
    ## (3) erase the invisibe regions in such a way that
    ##     at least the right side of the pick map remains
    ##     intact. NOTICE This task is accomplished using
    ##     the circle-mask procedures, which are attributes
    ##     of all patterns which belong to the sapatt class
    ## (3a) implant smask into a sapatt-pattern for operation
    fibr.mapp = smask.copy()
    smask1 = smask.copy()
    # fibr.save('test_smask1') # TEST
    
    ### Tests of killdisk methods TEST
    # fibr.mapp = 0.0*fibr.mapp + 1.0 # White everywhere    
    # fibr = killoutdisk( fibr, 200, 50, 100 )    
    # fibr.save('test_killdisk')

    ## (3b) Erase invisible regions. In doing so: sacrifice
    ##      the left side (negative x12) of the pattern but
    ##      leave the right side intact. The left side can
    ##      later be restored because of mirror symmetry
    killdisk( fibr, p1c, p3c, int(1/(wavel*del_s)) )
    killoutdisk( fibr, p1c, p3c, int(1/(wavel*del_s)) )
    smask2 = fibr.mapp.copy()
    # fibr.save('test_smask2') # TEST TODO

    # smask3 = img.mapp > 0.0    
    fibr.mapp = smask1 * smask2
    # fibr.save('test_smask') # TEST TODO
    
    ## (3d) Get back the "to-do-list" with all the s-values which
    ##      can be mapped (on the right side of the pattern)
    smask = fibr.mapp
    svalid = smask > 0 # All valid points (well; it's already identical to smask)
    svaltu = np.where( smask > 0 ) # as tuples: the valid pixels in s-space pattern
    # The structure of svaltu is special. The statement: 
    #   svaltu0, svaltu1 = svaltu 
    # splits it into two long arrays of 0-coordinate (Chinese, running down)
    # and 1-coordinate (running left to right). Important for cleanup.

    # *********************************************************************
    # * Map it
    # * Geometric mapping is accomplished by the Polanyi Equation
    # * cos(delta) = [s3/s - sin(theta) sin(beta)] / [cos(theta) cos(beta)]
    # * with   sin(theta) = (wavel/2)*s
    # * delta is the angle between the positive meridional axis and the
    # * pixel on the detector pattern, i.e. cos(delta) = p3/p
    # * 
    # * The radius on the detector is:   p = samp*tan(2*theta).
    # *
    # * Finally, the pixel coordinate p3 in vertical direction is
    # *     p3 = p * cos(delta)
    """
  We know approximately for which pixels in s-space we have intensities on 
  the detector image. So we can set up a reverse mapping rule: Given a pixel
  in s-space, from which point (p1,p3) on the detector does it receive its
  intensity? But before we apply the mapping rule, we have to test for each
  pixel, if the reverse mapping really ends inside the matrix of the detector
  image. While IDL handles this cleanup for the user, Python does not - and
  programming this is a challenge because of strange pythonic data types
  
  Notation: Most 2D arrays ("maps") cary a capital "M" in their name
    """
    # *********************************************************************

    # Make map of s-vector lengths in the valid mappable region
    Mslen = del_s * MEukl * svalid
    # print('Map it s-len: %6.3f .. %6.3f'%(np.min(Mslen[svaltu]), np.max(Mslen[svaltu])))
    # Make similar map of s3-values 
    ones = np.ones( fibr.width )
    grid = np.arange(fibr.height) - 0.5*(fibr.height - 1.0)
    # Because of Chinese addressing: arguments in "outer" swapped
    Ms3 = np.outer( grid, ones ) * np.double(del_s)*svalid
    # print('Map it s3   : %6.3f .. %6.3f'%(np.min(Ms3[svaltu]), np.max(Ms3[svaltu])))
    
    SinTheM = Mslen*wavel*np.double(0.5) # sin(theta) for each (s1,s3)
    CosTheM = np.sqrt(1 - SinTheM**2)    # cos(theta)

    # Make map of cosines of the polar angle delta on the detector
    #   delta = Pi/2 + alpha (alpha: elevation angle from equator)
    # For untilted fiber: cos(delta) = cos(phi)/cos(theta). 
    # With: phi: Polar angle on orientation sphere (after mapping)
    #   2*theta: scattering angle. Ms3/Mslen = cos(phi)
    CosDelM = np.zeros((p_len,p_len)) # preset empty matrix
    
    CosDelM[svaltu] = (Ms3[svaltu]/Mslen[svaltu] - SinTheM[svaltu]*np.sin(beta)) / \
        (CosTheM[svaltu]*np.cos(beta))

    # Generate the reflection-circle radius p on the detector related 
    # to the pixel in the future result that is defined by the position
    # inside the pmap
    Tan2TheM = SinTheM/CosTheM # well, now this is only tan(theta)
    # .. so make tan(2*theta) by application of the trigonometric
    # double-angle identity tan(2*theta)=2*tan(theta)/(1-tan(theta)**2)
    tt2 = Tan2TheM**2
    Tan2TheM = 2*Tan2TheM/(1 - tt2)
    tt2=0 # release memory
    # Compute the reflection circle radius on the detector in pixel units
    pmap = np.zeros((p_len,p_len)) # preset empty matrix
    pmap[svaltu] = samp*Tan2TheM[svaltu]
    # print('Det radii: %6.1f .. %6.1f'%(np.min(pmap[svaltu]), np.max(pmap[svaltu])))

    # Generate the p3-coordinate related to this pixel in the future result
    p3map = 0.0 * smask.copy()
    p3map[svaltu] = CosDelM[svaltu]*pmap[svaltu]
    print('Det p3: %6.1f .. %6.1f'%(np.min(p3map[svaltu]), np.max(p3map[svaltu])))

    # Compute the p1-coordinate of the pixel in p1map
    p1map = 0.0 * smask.copy()
    p1map[svaltu] = np.sqrt( pmap[svaltu]**2 - p3map[svaltu]**2 )
    print('Det p1: %6.1f .. %6.1f'%(np.min(p1map[svaltu]), np.max(p3map[svaltu])))
    # correct the sign
    p1map = -p1map
    
    # Change from coordinates relative to the center of the scattering pattern
    # to coordinates addressing the pattern matrix directly (origin top left)
    p1map = ( p1map+p_cen )*svalid # svalid is a mask (first cleanup step)
    p3map = ( p3map+p_cen )*svalid
    # Python cannot address matrix elements using floating point indices.
    # Moreover, if values in a matrix are rounded, the matrix loses its shape
    # and becomes a vector. So circumvent the problem by filling integer
    # indices into matrices of integer type ("int").
    p1int = np.zeros((p_len,p_len),int)
    p1int[svaltu] = np.rint(p1map[svaltu])
    p3int = np.zeros((p_len,p_len),int)
    p3int[svaltu] = np.rint(p3map[svaltu])
    # Now still some of the coordinate pairs p1int[svaltu],p3int[svaltu]
    # in the svaltu tuple point out of the measured pattern img. Remove
    # these dangling pointers!

    pup = p_len - 1
    # Because tuples are immutable: work with lists
    # svaltu is a tuple which contains a long list of positions in the future
    # result pattern, most of which can be related to a point on the measured
    # detector pattern.
    # Separate svaltu into two arrays:
    # svaltu0: first coordinate, running down in the s-map
    # svaltu1: second coordinate, running right in the s-map
    svaltu0, svaltu1 = svaltu
    # Make svaltu0 and svaltu1 "mutable" by converting them into lists.
    # Then from svalli0 and svalli1 we can remove bad mappings.
    svalli0, svalli1 = list(svaltu0), list(svaltu1)
    # upli is the number of mappings (one for each pixel)
    upli = len(svalli0)
    print('Before cleanup: %i pixel-to-pixel mappings'%upli)
    for i in range(upli):
        # Proceed in reverse order - otherwise already the first purge
        # saws off the branch BEHIND us on which we climb upward.
        k = upli - i - 1 # in reverse order
        p1 = p1int[svalli0[k],svalli1[k]] # extract p1 coordinate of the mapping
        p3 = p3int[svalli0[k],svalli1[k]] # same for p3
        if (p1 < 0) or (p3 < 0) or (p1 > pup) or (p3 > pup):
            # Mapping runs out of the detector pattern: Cleanup necessary
            p1int[svalli0[k],svalli1[k]] = 0 # Just for fun: point at upper left
            p3int[svalli0[k],svalli1[k]] = 0 # corner of the original pattern
            svalli0.pop(k) # remove this item from lists
            svalli1.pop(k) # remove this item from lists
            pass
        pass
    # Back to tuple
    svaltu0, svaltu1 = np.array(svalli0), np.array(svalli1)
    svaltu = tuple((svaltu0, svaltu1))
    upli = len(svalli0)
    print('After cleanup:  %i pixel-to-pixel mappings'%upli)
     
    # Build the Jacobian for intensity correction
    Mcos2The = 1.0 - (wavel*Mslen)**2/2
    # The fiber-geometry Jacobian
    jacmap = (wavel*samp)**2/Mcos2The**3

    # Geometric mapping (img) and intensity mapping (jacmap)
    afac = Mcos2The # Absorption factor
    qp = 0.98 # Polarization quality for synchrotron beamline (DESY, A2)
    
    # Map of s1 values for polarization factor
    Ms1 = np.outer( ones, grid ) * np.double(del_s)*svalid
    fibr.mapp = Ms1
    # fibr.save('test_s1')

    # The polarization factor -- Equation:
    # pfac = 0.5*[(1-q)*cos(phi)**2 + (1+q)*sin(phi)**2]
    # With q=0.98 the meridional term (cos) is very small and a division
    # pushes the intensity extremely up. This is unreasonable.
    pfac = np.zeros((p_len,p_len))
    pfac[svaltu] = 0.5*( (1-qp)*(Ms3[svaltu]/Mslen[svaltu])**2 + \
        (1+qp)*(Ms1[svaltu]/Mslen[svaltu])**2 )
    # fibr.mapp = pfac
    # fibr.save('test_polarisation') # TEST

    # Python expects indices to be strictly integer.
    p1int = np.zeros((p_len,p_len),int)
    p1int[svaltu] = np.rint(p1map[svaltu])
    p3int = np.zeros((p_len,p_len),int)
    p3int[svaltu] = np.rint(p3map[svaltu])

    # FINALLY, THE MAPPING IS ONLY A SINGLE LINE OF CODE:
    # WARNING Chinese addressing with img.mapp!!!
    fibr.mapp[svaltu] = img.mapp[p3int[svaltu], p1int[svaltu]] * jacmap[svaltu] / afac[svaltu]
    # fibr.mapp[svaltu] = fibr.mapp[svaltu] / pfac[svaltu]
    # TODO The effect of the polarization factor pfac is strange and much too strong.
    #      probably the quality q is incorrect.

    # Mirror to the left side of the pattern. Transpose and rotate
    # WARNING Chinese addressing requires 270° rotation
    fibr.mapp = fibr.mapp + np.rot90(fibr.mapp.T,3)
    # fibr is the result. Store it in a file!
    filebits = thefile.split("_")
    # Replace the first file bit with a special indicator string. Here:
    filebits[0] = "hars" # Harmonized in s-space
    outfile = "_".join(filebits)
    fibr.save(outfile)
    return ''
# END map_it
# END script procedures

# BEGIN Main:
err = map_it()
if err != '':
    # Error treatment of cmd2 version 0.7.8 (source: savcmd2.py)
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()
