#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types

'''wf_multiband.py: 
  
SYNTAX:   @@wf_multiband,<controlfile>,<framefile>[.sjs]
EXAMPLES: @@wf_multiband,zzbands.dat,hart_qa15_0000

     REQUIREMENTS:
        WAXS Faserstreubild, das auf die (s_12,s_3)-Ebene
        abgebildet ist. Quadratisch, zentriert und harmonisch.
        Wenn das Attribut muba_a0 existiert, wird die Arbeit 
        verweigert. Abhilfe: patt.dela('mula_a0')
        
        A control file (e.g. zzbands.dat) containing something like:

        0.305 0.605 10
        # Control file for wf_multiband.py meaning:
        # Divide the range between 3.05 und 6.05 Å into 10 bands
        # and carry out a multipole texture analysis in each band
    
    METHOD
        Each band is projected individually onto the orientation sphere
        for a multipole analysis. The results are returned in 1D arrays
        which contain as many elements as there are bands. Each array
        either specifies the band itself or one of the resulting parameters.
        The arrays are stored in the scattering pattern as attributes.        
        muba_db[:] : smallest d-spacing (1/se) of the band
        muba_de[:] : biggest d-spacing (1/sb) of the band
        muba_It[:] : total intensity scattered into the spherical shell
                     defined by the band
        muba_a0[:] : 0th Legendre coefficient (Norm on the orientation sphere)
        muba_fh[:] : a2/a0. Hermans' orientation parameter (not yet transformed)
        muba_f4[:] : a4/a0. Quadrupole moment (for curiosity)
    
'''

## Global environmental parameters
scale_smap = 1.0 # Scaling factor for nice plotting of hops curves
scale_iht = 1e-5 # Scaling factor für I_htot-curve
scale_a0  = 1e-5 # Skaling for a0 in the attribut which makes  a0 = I_htot

## Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def Dist( quad ):
    '''
    Die Dist() Funktion vom PVWAVE - aber nur ein Quadrant
    (Unten rechts in einem harmonisierten Bild)
    Dist(4) ergibt klassisch:
    [[0.         1.         2.         3.        ]
    [1.         1.41421356 2.23606798 3.16227766]
    [2.         2.23606798 2.82842712 3.60555128]
    [3.         3.16227766 3.60555128 4.24264069]]
    Beim Streubild starten die Pixelpositionen mit 
    0.5 in 1er-Schritten, also mit der Streukonvention:
    [[0.70710678 1.58113883 2.54950976 3.53553391]
    [1.58113883 2.12132034 2.91547595 3.80788655]
    [2.54950976 2.91547595 3.53553391 4.30116263]
    [3.53553391 3.80788655 4.30116263 4.94974747]]
    '''
    # axis = np.linspace( 0, quad, quad ) # NOTICE Klassisch
    axis = np.linspace( 0.5, quad-0.5, quad ) #NOTICE Streu-Konvention
    return np.sqrt(axis**2 + axis[:,np.newaxis]**2)

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def doit():
    '''doit does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))
    
    arglist = argexpand( 2 ) # expand to 2 arguments
    
    ## (1) The control file
    bandcontrol = arglist[0]
    try:
        ctrlfil = open( bandcontrol )
    except:
        return 'Cannot open the band control file "%s"'%thecontrol
    bandlin = ctrlfil.readline().split() # space as separator
    ctrlfil.close()
    dlow  = float(bandlin[0])
    dhigh = float(bandlin[1])
    nb    =   int(bandlin[2])
    # The bands are assumed equidistant in d, but not in s
    print('Going to split d-range [%5.2f nm .. %5.2f nm] into %i bands'%(dlow,dhigh,nb))
    delta =(dhigh-dlow)/nb
    dlims = dhigh - np.arange(nb+1)*delta # reverse order
    slims = 1.0/dlims # normal order in s
    
    
    ## (2) The image file
    thefile = arglist[1]
    patt = sp.Patt()
    if patt.load( thefile ):
        return 'Loading error'
    if patt.width != patt.height:
        return 'Pattern must be quadratic'
    if not patt.hasa('center'):
        return 'Center attribute missing'
    cenx, ceny = patt.geta('center')
    if not patt.hasa('Compressible1/4'): # Attr missing
        return 'Pattern not yet harmonized'
    if not patt.geta('Compressible1/4'): # Wrong Attr value
        return 'Pattern shows no harmony'
    #if patt.hasa('muba_a0'):
    #    return 'Pattern has already an attribute "muba_a0"'
    # relaxation time of the frame
    # relaxtime = patt.geta("relaxtime") NOTICE required?

    if patt.vaxras[:-1] == [0.172,0.172]:
        return 'Pattern is still in detector coordinates'

    # Make an s-distance quadrant of pixel centers
    deltas = patt.vaxras[0]
    QsDists = Dist(cenx)*deltas
    # Work with a quadrant of intensities
    QI = patt.mapp[ceny:,cenx:]
    # Mask of good pixels
    maskI = QI > 0.0
    # Check, if s-limits are inside the image

    # last valid pixel position on equator (remember Chinese addressing)
    maskI_s = maskI*QsDists # s-values of good pixels
    # Outer edge of outermost good pixel on equator
    seqmax = np.max(maskI_s[0,:]) + 0.5*deltas 
    if seqmax < slims[-1]:
        dh=1.0/seqmax
        print('*** Edit bandcontrol file %s. Make d_min > %6.4f'%(bandcontrol,dh))
        return 'Pattern does not cover multibands     ^^^.'
        
    ### BEGIN Über alle Bänder
    ##  muba: Arrays for storage of the results before they become
    ##        attributes
    bands = len(slims)-1
    muba_db = np.empty(bands) # User friendly: band start in r-space
    muba_de = np.empty(bands)
    muba_It = np.empty(bands)
    muba_a0 = np.empty(bands)
    muba_fh = np.empty(bands)
    muba_f4 = np.empty(bands)
    for iband in range(len(slims)-1):
        # Decreasing in s means increasing in d-spacing
        sb = slims[-(iband+2)]
        se = slims[-(iband+1)]
        # print('%2i: %6.4f .. %6.4f'%(i,sb,se))
        masklo = QsDists <= se # =0 where too far out
        maskhi = QsDists >  sb # =0 where too close to origin
        dmask  = masklo*maskhi*maskI # Pixels carrying scattering in quadrant
        
        # BEGIN to project the band
        goodpix = np.where(dmask==1) # only in the actual band
        # goodpix is an array with two columns filled with the y-
        # and the x-indices of the good pixels
        NumGood = len( goodpix[0] )
        rowi,coli = goodpix # Split indices for easy access
        '''
        The good pixels (goodpix) are not yet sorted, but for sorting
        their (x,y) coordinates must be entered into a 2D array which
        then is sorted ONLY by increasing x-values (so that the y-values
        are shifted adhering to their x-values)
        '''
        
        funcol = np.empty([NumGood,2])
        for i in range(NumGood):
            s3 = rowi[i]
            s12 = coli[i]
            fval = QI[s3,s12]
            # Polar angle away from the s3-axis
            phi = 180*np.arctan((s12+0.5)/(s3+0.5))/np.pi
            # print( 'Wert:%7.3f Winkel: %5.2f° (%3i,%3i)'%(fval,phi,s12,s3) )
            funcol[i] = [phi,fval]
            pass
        self.funcol = funcol

        # Sort the pairs by increasing polar angle (column 0)
        kurve = funcol[funcol[:,0].argsort()]
        # Apply factor for handy plotting
        kurve[:,1] = kurve[:,1]*scale_smap
        
        # Save halo on the pole figure with resolution of 1°.
        # phi range: 0.5 .. 189.5; symmetrical around the equator
        dots = 180
        halocur = np.zeros([dots,2])

        accu = 0.0
        count = 0
        slotphi = 0.5
        for i in range(NumGood):
            # The NumGood in kurve are sorted
            if kurve[i,0] <= slotphi+0.5:
                accu = accu + kurve[i,1] # Integrate by accumulation
            else:
                # Where to enter?
                aufi = int(slotphi) # ... upwards
                abi  = 179-aufi # ... downwards
                halocur[aufi,0] = slotphi
                halocur[aufi,1] = accu
                halocur[abi,0] = 180-slotphi
                halocur[abi,1] = accu
                accu = kurve[i,1]
                slotphi = slotphi + 1.0 # Increment angle by 1°
        # Enter last pair of values

        if slotphi < 90.0: # Extra treatment for last point
            aufi = int(slotphi) # upwards
            abi  = 179-aufi # downwards
            halocur[aufi,0] = slotphi
            halocur[aufi,1] = accu
            halocur[abi,0] = 180-slotphi
            halocur[abi,1] = accu
            
        thehalo = self.ops[1] # pole figure curve in SASkia
        thehalo.clean()
        thehalo.x[:dots] = halocur[:,0] # Polar angle phi
        thehalo.y[:dots] = halocur[:,1] # pole figure intensity
        thehalo.n = dots
        
        # Remove spikes form detector structure 
        smo = thehalo._spline( 0.5, 0 )
        thehalo.y[:dots] = smo.y[:dots] # transfer smoothed values
        # NOTICE smo.dy[:dots] would be the derivative
        # END projection of band onto the orientation sphere (curve: thehalo)
        
        # BEGIN WARNING Fix artifacts for phi < 10°
        # These values increase, because the wedge fit returns a surface which
        # is too high. Next time: smooth before mapping, map also the smoothed
        # pattern, compute the extrapolation surface from the smoothed data
        # and use this to fill the unsmoothed pattern in the wedge.
        lim = 10
        fixwert = thehalo.y[lim]
        thehalo.y[:lim-1] = fixwert
        thehalo.y[dots-lim+1:dots] = fixwert
        # END WARNING Fixing of edge values

        
        # BEGIN Multipole expansion
        n2e = dots/2 # from north pole to equator
        vphi = halocur[:n2e,0] # Vector of phi
        vray = halocur[:n2e,1] # Vector of x-ray intensity
        # From degrees to radians
        vphi = vphi*np.pi/180
        vsin = np.sin(vphi)
        vcos = np.cos(vphi)
        vcos2= vcos*vcos
        vraysin = vray*vsin # Present in each integral
        # Coefficient a0 with Legendre-Polynomial P0=1
        # NOTICE Integration generally only up to Pi/2 - This cancels pre-factor 0.5
        a0 = np.sum(vraysin)
        # Coefficient a2 with Legendre-Polynomial P2
        # NOTICE Pre-factor 5 omitted (as is common in scattering)
        a2 = np.sum(vraysin*0.5*(3*vcos2 - 1))
        fh = a2/a0 # Hermans orientation factor
        # Coefficient a4 with Legendre-Polynomial P4 with pre-factor
        a4 = (9.0/4)*np.sum(vraysin*(35*vcos**4-30*vcos2+3)/8)
        f4 = a4/a0
        # END Multipole expansion
        a0 = a0*scale_a0
        
        # BEGIN Compute I_htot: Total intensity of the halo
        s12vec = np.linspace(0.5,cenx-0.5,cenx) # in units of pixel
        s12vec = s12vec*patt.vaxras[0] # In nm^-1 (comes more handy)
        Qs12 = np.outer(np.ones(cenx),s12vec)    
        # Quadrant of croissant intensities ( I*s12 ) 2*Pi omitted
        Qring = QI*Qs12*dmask
        iht = np.sum( Qring )*scale_iht
        # END I_htot Computation
        # Store results in muba-vectors
        dbeg = 1/se
        dend = 1/sb
        muba_db[iband] = dbeg # reverse order
        muba_de[iband] = dend # reverse order
        muba_It[iband] = iht
        muba_a0[iband] = a0
        # For chain dir: fhm = - 2*fh
        muba_fh[iband] = fh # Hermans orientation parameter (NOT in chain dir)
        muba_f4[iband] = f4

        # NOTICE Log to screen
        #print( "%2i [%5.3f,%5.3f]: Iht =%7.4f a0 =%7.4f, fh = %7.4f, f4 =%7.4f"%(iband,dbeg,dend,iht,a0,fh,f4) )    
    ### END loop all bands
    
    # Book keeping

    # Store results as attributes to the pattern
    patt.puta('muba_db',muba_db)
    patt.puta('muba_de',muba_de)
    patt.puta('muba_It',muba_It)
    patt.puta('muba_a0',muba_a0)
    patt.puta('muba_fh',muba_fh)
    patt.puta('muba_f4',muba_f4)
    patt.logentry('# muba texture in d-range [%5.2f nm .. %5.2f nm] split into %i bands'%(dlow,dhigh,nb))

    # Save the pattern back to disk
    patt.save( thefile )

    
    return ''
#END doit()

# END script procedures

# Main:
err = doit()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()



