#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types

'''wf_polsana.py: Polfigur-Analyse eine WAXS-Streubilds mit
    Fasersymmetrie nach der Abbildung in den s-Raum (nach der
    Ausführung von wf_mapxxxx). Das Bild muss harmonisiert und
    quadratisch sein. Die Faserachse (s3) ist senkrecht). Die 
    blinden Zwickel am Meridian sind durch Extrapolation gefüllt.

    ARBEITSWEISE
        Im echten s-raum ist es einfach: Quadrant picken,
        Maske setzen, auf [s_s .. s_e] einschränken (patdist),
        mit s12 multiplizieren, zu jedem Pixel phi bestimmen,
        die Kranz-Intensitäten in die Ausgabe-Kurve hineinzählen
        und zum Schluß für die Polfigur bei jedem phi wieder auf
        dem Kranz der Einheitskugel (2*Pi*sin(phi)) verteilen.
    
SYNTAX:   @@wf_polsana,<framefile>[.sjs]
EXAMPLES: @@wf_polsana,hars_qa15_0000
    
    <framefile>:  Das Faserstreubild in der (s12,s3)-Ebene.
                  WARNING Python addresses 2D arrays the Chinese way:
                          I_map( down, right )
    Output: Streukurven und Parameterwerte
'''

## TODO Specify global setup parameters ALWAYS TODO
valsb = 1.25 # value of s_begin in nm^{-1} upper limit in r: 8 Å
valse = 3.28 # value of   s_end in nm^{-1} outmost valid pixel
scale_smap = 1e-3 # Skalierung, die für leicht plotbare "hops"-Kurven sorgt
scale_iht = 1e-5 # Skalierung für I_htot-Kurve
scale_a0  = 1.0/44 # Skalierung für Ausgabe der a0-Kurve, die a0 = I_htot macht

## Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def Dist( quad ):
    '''
    Die Dist() Funktion vom PVWAVE - aber nur ein Quadrant
    (Unten rechts in einem harmonisierten Bild)
    Dist(4) ergibt klassisch:
    [[0.         1.         2.         3.        ]
    [1.         1.41421356 2.23606798 3.16227766]
    [2.         2.23606798 2.82842712 3.60555128]
    [3.         3.16227766 3.60555128 4.24264069]]
    Beim Streubild starten die Pixelpositionen mit 
    0.5 in 1er-Schritten, also mit der Streukonvention:
    [[0.70710678 1.58113883 2.54950976 3.53553391]
    [1.58113883 2.12132034 2.91547595 3.80788655]
    [2.54950976 2.91547595 3.53553391 4.30116263]
    [3.53553391 3.80788655 4.30116263 4.94974747]]
    '''
    # axis = np.linspace( 0, quad-1, quad ) # NOTICE Klassisch
    axis = np.linspace( 0.5, quad-0.5, quad ) #NOTICE Streu-Konvention
    return np.sqrt(axis**2 + axis[:,np.newaxis]**2)

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def doit():
    '''doit does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))
    
    arglist = argexpand( 2 ) # expand to 2 arguments
    
    ## (1) The image file
    thefile = arglist[0]
    b = sp.Patt()
    if b.load( thefile ):
        return 'Loading error'
    if b.width != b.height:
        return 'Pattern must be quadratic'
    if not b.hasa('center'):
        return 'Center attribute missing'
    cenx, ceny = b.geta('center')
    if not b.hasa('Compressible1/4'): # Attr missing
        return 'Pattern not yet harmonized'
    if not b.geta('Compressible1/4'): # Wrong Attr value
        return 'Pattern shows no harmony'
    # relaxation time of the frame
    relaxtime = b.geta("relaxtime")

    # WARNING WARNING WARNING
    # The error in wf_maps2008 has been corrected. There only the first
    # of all patterns in a directory got the correct s-pixel-size.
    if b.vaxras[0] == 0.172: # still detector because of error in wf_maps2008?
        b.vaxras[:-1] = 0.01032, 0.01032 # size of s-mapped pixels
        # Enforce pixel size of August 2020 mapping
    # WARNING WARNING WARNING
    
    # NOTICE CONVENTION: This program works on scattering patterns
    #                    with fiber symmetry which are mapped on the
    #   (s12,s3) plane of the reciprocal space. Their name has a
    #   special "flavor"-Prefix which is followed by the frame designation
    #   The common flavor is "hars" (harmonized in s-space). The curve file
    # names which will be generated here shall no longer carry the flavor
    # anymore, but only keep the frame designation in its name.
    thebody = thefile.split('.')
    thebits = thebody[0].split('_')
    theframe = '_'.join(thebits[1:])

    # BEGIN Projektion Halo-Ring [raddn .. radup] auf die Lagenkugel    
    # China-Quadrant der Intensitätsdaten
    quad = b.mapp[ceny:,cenx:]
    # China-Quadrant mit den Abständen jedes Pixels vom Mittelpunkt
    Vals = Dist( cenx )*b.vaxras[0] # ... wird s durch *b.vaxras

    dmask1 = Vals < valse # =0, wo zu weit außen
    dmask2 = Vals > valsb # =0, wo zu weit innen
    dmask3 = quad > 0.0 # =0 bei blinden Pixeln
    dmask  = dmask1*dmask2*dmask3 # Strahlende im Viertelkreis
    
    if np.sum(dmask) < 10000:
        return 'Less than 100 x 100 pixels to project -- this is strange'
    
    goodpix = np.where(dmask==1) # Nur im Halo-Band
    # goodpix ist ein Feld mit zwei Spalten, in denen die y-
    # und x-Indices der guten Pixel stehen.
    NumGood = len( goodpix[0] ) # Anzahl der Guten
    rowi,coli = goodpix # Indices für Zugriff gespalten
    # Die goodpix-Punkte sind nicht sortiert, aber zum Sortieren
    # müssen deren (x,y)-Koordinaten in ein 2D Feld eingetragen
    # und dann NUR nach den x-Werten sortiert werden (so dass
    # die y-Werte an ihren x-Werten haftend verschoben werden
    
    funcol = np.empty([NumGood,2])
    for i in range(NumGood):
        s3 = rowi[i]
        s12 = coli[i]
        fval = quad[s3,s12]
        # Polarwinkel weg von s3-Achse
        phi = 180*np.arctan((s12+0.5)/(s3+0.5))/np.pi
        # print( 'Wert:%7.3f Winkel: %5.2f° (%3i,%3i)'%(fval,phi,s12,s3) )
        funcol[i] = [phi,fval]
        pass
    self.funcol = funcol

    # Die Wertepaare aufsteigend sortieren.
    # nach Polarwinkel (Spalte 0)
    kurve = funcol[funcol[:,0].argsort()]
    # Faktor für leichtes Plotten
    kurve[:,1] = kurve[:,1]*scale_smap
    
    # Halo auf der Polfigur mit Auflösung von 1° speichern
    # phi 0.5 bis phi 189.5; Symmetrie am Äquator;
    dots = 180 # Punktzahl
    halocur = np.zeros([dots,2])

    accu = 0.0
    count = 0
    slotphi = 0.5
    for i in range(NumGood):
        # Die NumGood sind in kurve sortiert
        if kurve[i,0] <= slotphi+0.5:
            accu = accu + kurve[i,1] # Funktionswert akkumulieren
        else:
            # Wo eintragen?
            aufi = int(slotphi) # ... aufwärts
            abi  = 179-aufi # ... abwärts
            halocur[aufi,0] = slotphi
            halocur[aufi,1] = accu
            halocur[abi,0] = 180-slotphi
            halocur[abi,1] = accu
            accu = kurve[i,1]
            slotphi = slotphi + 1.0 # Winkel um 1° rauf
    # Letztes Wertepaar eintragen

    if slotphi < 90.0: # Extra treatment for last point
        aufi = int(slotphi) # Aufwärts eintragen
        abi  = 179-aufi # Abwärts eintragen
        halocur[aufi,0] = slotphi
        halocur[aufi,1] = accu
        halocur[abi,0] = 180-slotphi
        halocur[abi,1] = accu
        
    thehalo = self.ops[1] # pole figure of halo in saskia
    thehalo.clean()
    thehalo.x[:dots] = halocur[:,0] # Polar angle phi
    thehalo.y[:dots] = halocur[:,1] # pole figure intensity
    thehalo.n = dots
    
    # Spikes von der Gitterstruktur des PILATUS-Detektors
    # wegglätten
    smo = thehalo._spline( 0.5, 0 )
    thehalo.y[:dots] = smo.y[:dots] # Funktionswerte übertragen
    # NOTICE smo.dy[:dots] wäre die 1. Ableitung
    # END Projektion Halo-Ring auf die Lagenkugel (Kurve thehalo)
    
    # BEGIN WARNING Künstlich steigende Randwerte für phi < 10° fixieren
    # Die Randwerte steigen, weil der Zwickelfit mit radb zu hoch steigt.
    lim = 10
    fixwert = thehalo.y[lim]
    thehalo.y[:lim-1] = fixwert
    thehalo.y[dots-lim+1:dots] = fixwert
    # END WARNING Randwertfixierung

    
    # BEGIN Multipole expansion
    n2e = dots/2 # from north pole to equator
    vphi = halocur[:n2e,0] # Vector of phi
    vray = halocur[:n2e,1] # Vector of x-ray intensity
    # From degrees to radians
    vphi = vphi*np.pi/180
    vsin = np.sin(vphi)
    vcos = np.cos(vphi)
    vcos2= vcos*vcos
    vraysin = vray*vsin # Present in each integral
    # Coefficient a0 with Legendre-Polynomial P0=1
    # NOTICE Integration generally only up to Pi/2 - This cancels pre-factor 0.5
    a0 = np.sum(vraysin)
    # Coefficient a2 with Legendre-Polynomial P2
    # NOTICE Pre-factor 5 omitted (as is common in scattering)
    a2 = np.sum(vraysin*0.5*(3*vcos2 - 1))
    fh = a2/a0 # Hermans orientation factor
    # Coefficient a4 with Legendre-Polynomial P4 with pre-factpr
    a4 = (9.0/4)*np.sum(vraysin*(35*vcos**4-30*vcos2+3)/8)
    f4 = a4/a0
    # END Multipole expansion
    
    # BEGIN Compute I_htot: Total intensity of the halo
    s12vec = np.linspace(0.5,cenx-0.5,cenx) # in units of pixel
    s12vec = s12vec*b.vaxras[0] # In nm^-1 (comes more handy)
    s12quad = np.outer(np.ones(cenx),s12vec)    
    # Quadrant of croissant intensities ( I*s12 ) 2*Pi omitted
    ringquad = quad*s12quad*dmask
    iht = np.sum( ringquad )*scale_iht
    # END I_htot Berechnung

    # Log to screen
    print( "Iht =%7.4f a0 =%7.4f, fh = %7.4f, f4 =%7.4f"%(iht,a0,fh,f4) )    

    thehalo.puta("units1","$^\\\\circ$")
    # units2 bleibt a.u.
    thehalo.puta("axis1","$\\\\varphi$")
    thehalo.puta("axis2","${{I}}_o( \\\\varphi )$")
    
    prefix = "hops_" #NOTICE halo on pole figure from s-space
    hpffile = prefix + theframe

    # Origin and bookkeeping
    thehalo.headers.append('# I_halo(phi) on pole figure of "%s"' %thefile)
    thehalo.puta("fromfile",hpffile)
    thehalo.puta("fn",b.geta("fn")) # original data
    thehalo.puta("relaxtime",relaxtime) # WARNING related to experiment

    # New parameters
    multi = [iht, a0*scale_a0, fh, f4] # I_ht and all multipole parameters into an array
    thehalo.puta("multipole",multi)
        
    # Save curve and parameters
    # Make a new file name based on the origin of this curve
    # NOTICE log to screen
    print('Writing the curve to %s [.dat]'%hpffile)
    thehalo.writeascii( hpffile, True ) # Allow overwriting

    return ''
#END doit()

# END script procedures

# Main:
err = doit()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()
