#!/usr/bin/env python
# -*- coding: utf-8 -*-
# samonotiff.py
# Almut Stribeck almut@stribeck.de

'''Suggestion:
    import monotiff as mt
    m, t = mt.tiffread( 'myfile.tif' )
        After saving your own changes: reload(mt)
    import Scat2 as sp
    a = sp.ScatPatt( m, ( t["x_micron"]/1000, 'mm' ) )
    a.puta( 'title', t["document_name"] )
    a.puta( 'timestamp', t["time_stamp"] ] )
       ...
       and whatever property should be documented
       under which name. Probably a second file with
       environmental data (like ".fio") must be
       consulted to retrieve all the parameters for
       an evaluation of the scattering pattern

    # reading data from the pattern
    a.geta('timestamp') # read the time stamp
    a.width # the width of the pattern in pixels
    a.mapp # 2D pattern map of numpy.float32
'''
from __future__ import print_function # coherent "print" for Python 2 and 3
import numpy, struct

def tiffread( filename ):
    '''
    Simple general TIFF reader for grayscale data.

    Usage: import monotiff as mt
           m, t = mt.tiffread('myfile.tif')

           m : the image as a matrix numpy.float32
           t : a dictionary containing essential
               tags and the corresponding values.
               Additionally, the virtual keys
               "x_micron" and "y_micron" open
               access to the size of each pixel
               in micrometers -- if the tags 282,
               283 and 296 have been obeyed by the
               programmer of the tiff-writer.

    Reason: I have read TIFF files in many programming languages
    using my own reader. Now with python there are many complex
    modules that permit to read TIFFs, but all these modules share
    the same error: from a strictly positive gray-scale image they
    can produce an image with negative intensities whenever the
    highest bit in a 4-byte-integer is set (int32 vs. uint32
    interpretation - i.e. data are considered signed 4-byte integers
    instead of unsigned ones). Moreover, I hate to use these
    monstrous al-rounder modules that "document themselves" but
    are so complex that it takes weeks to get an overview.

    Similar to the python image library, I do no longer try to
    unravel all the tags that some programmer has dumped into the
    file, because there are so many inconsistent TIFF files around.
    This is especially the case for TIFFs generated by the PILATUS
    detector.

    I only read the tags that I consider important.

    WARNING: Pilatus detector files carry resolution information
    both in an "image_description" string tag, and in the tags
    "x_resolution" and "y_resolution". The two sets of information
    are inconsistent. Moreover, the "resolution_unit"-tag is not
    even set. In this reader I have preset the tag with a default
    value of 2 so that it reflects the intention of the programmer
    who wrote the TIFF writer. Nevertheless, there remain some
    inconsistencies that I do not want to heal here, because they
    would spoil the reader.
      Until the guys have fixed it one should tune the values in
    x_micron and y_micron and make them 172 -- in the outer space.
      The pixels that carry extreme spikes in every Pilatus file
    should be set to zero -- as well in the outer space.
      If all those pixels are written with the top bit set, the
    threshold is hex(80000000) = 2147483648.

    almut@stribeck.de
    '''
    if not isinstance( filename, str ):
        raise AttributeError("filename must be a string. Quote it!")
    lun = open( filename, 'rb' )
    # Read TIFF stamp as string from image file header
    stamp = lun.read(2)
    if not (stamp in ('II', 'MM')):
        raise ValueError("First 2 bytes do not comply to TIFF definition")

 # # remember byteorder as required by struct
    if stamp == 'II':
        # Little-Endian (Intel CPU)
        byteorder = '<'
    else:
        # Big-Endian (Motorola CPU)
        byteorder = '>'
    # Extract TIFF version stored in an uint16 (Type: "H" -- half)
    version, = struct.unpack(byteorder+'H',lun.read(2))
    if version == 43:
        raise ValueError("Big TIFF format not supported")
    if version != 42: # Douglas Adams': "The answer is 42!"
        raise ValueError("TIFF: Wrong answer, Douglas -- this is no TIFF file")
    # Get pointer to the 1st image file directory uint32
    ifd_p, = struct.unpack(byteorder+'L',lun.read(4))
    # Place file pointer there
    lun.seek(ifd_p)
    # Retrieve the number of directory entries
    nifde, = struct.unpack(byteorder+'H',lun.read(2))
    # Read all directory entries into a list
    ifd = []
    for i in range(nifde):
        ifde = lun.read(12) # Each entry 12 bytes
        ifd.append( ifde )

    # Dictionary of the tags that shall be considered and their meaning
    esstag = {256:"image_width", 257:"image_height", 258:"bits_per_sample",
              259:"compression", 269:"document_name", 273:"strip_offset",
              278:"rows_per_strip", 282:"x_resolution", 283:"y_resolution",
              284:"planar_config", 296:"resolution_unit", 306:"time_stamp"}
    '''List of corresponding default values
    image_width = None
    image_height = None
    bits_per_sample = 1
    compression = 1     # no compression
    document_name, = filename.rpartition('.') # until the last dot
    strip_offset = None
    rows_per_strip = None
    x_resolution = None
    y_resolution = None
    planar_config = 1
    resolution_unit = 2 # in inch 3 is "cm" and 1 is "relative"
    time_stamp = ""

    '''
    essdef = [ None, None, 1, 1, filename.rpartition('.')[0],
               None, None, None, None, 1, 2, " " ]
    # tagvals: to be filled and returned in the result
    # essdef values will be overwritten by extracted values
    tagvals = dict(zip( esstag.values(), essdef ))

    # The lengths of the TIFF types -- with a leading entry 0
    # to compensate for the indexing of python tuples, arrays ...
    typlen = (0,1,1,2,4,8,1,1,2,4,8,4,8)
    for i in range(nifde):
        tag, = struct.unpack(byteorder+'H',ifd[i][0:2]) # stays below index 2
        if tag in esstag.keys(): # Don't bother about unwanted tags
            # unravel the related value
            typ, = struct.unpack(byteorder+'H',ifd[i][2:4])
            cnt, = struct.unpack(byteorder+'L',ifd[i][4:8])
            # number of bytes related to the value in the file
            nbytes = cnt*typlen[typ]
            # if this is higher than 4, the following 4 bytes are
            # a pointer into the file. Else they represent the value
            # itself
            if nbytes > 4:
                offset, = struct.unpack(byteorder+'L',ifd[i][8:12])
                lun.seek( offset )
                val = lun.read( nbytes )
            else:
                val = ifd[i][8:12]
            # Convert val
            if typ == 2:
                # string
                if cnt == 0:
                    val = " "
                else:
                    val = val[0:cnt]
                    # catch zero padded data
                    while val[cnt-1] == "\x00":
                        cnt = cnt - 1
                    val = val[0:cnt]
            elif typ == 3:
                # short integer
                val, = struct.unpack(byteorder+'H',val[0:2])
            elif typ == 4:
                # long integer
                val, = struct.unpack(byteorder+'L',val[0:4])
            elif typ == 5:
                # rational number
                nom, = struct.unpack(byteorder+'L',val[0:4])
                den, = struct.unpack(byteorder+'L',val[4:8])
                val = 1.0*nom/den
            # overwrite value
            tagvals[esstag[tag]] = val
        # tag was in esstag
    # end of for loop

    # Test the tags for consistency
    if tagvals["compression"] != 1:
        raise TypeError("TIFF file appears compressed")
    if tagvals["planar_config"] != 1:
        raise TypeError("Not a simple planar configuration in TIFF")
    if tagvals["rows_per_strip"] != tagvals["image_height"]:
        raise TypeError("Multi-strip format unsupported")

    bps = tagvals["bits_per_sample"]
    if (bps != 8) and (bps != 16) and (bps != 32):
        print('The intensities pretend to be stored with ',bps,' bits per sample')
        raise TypeError("Supported are only 8, 16 or 32 bits per pixel")

    # compute pixel size from resolution in lines per ...
    ru = tagvals["resolution_unit"]
    xr = tagvals["x_resolution"]
    yr = tagvals["y_resolution"]
    if (xr != None) and (yr != None):
        if ru == 2:
            # lines per inch
            xpix = 25400.0/xr
            ypix = 25400.0/yr
        if ru == 3:
            # lines per cm
            xpix = 10000.0/xr
            ypix = 10000.0/yr
        # Add information about pixel extension in units of micrometers
        tagvals.update({"x_micron":xpix,"y_micron":ypix})
    else:
        print("x_resolution or y_resolution not found in the tags")

    w = tagvals["image_width"]
    h = tagvals["image_height"]

    offset = tagvals["strip_offset"]
    lun.seek( offset ) # Start of the data array (integer)
    if bps == 8: # The standard grayscale TIFF with 8 bit resolution
        mat = numpy.fromfile( lun, byteorder+"u1", w*h, )
    if bps == 16: # GEL files, MAR detector files: 16 bit resolution
        mat = numpy.fromfile( lun, byteorder+"u2", w*h, )
    elif bps == 32: # PILATUS detector files: 32 bit resolution
        mat = numpy.fromfile( lun, byteorder+"u4", w*h )
    # TIFF and numpy arrays have height and width swapped.
    mat = mat.reshape( h, w )
    mat = numpy.float32(mat).copy()
    # Correct swapping also in the tags
    tagvals["image_width"]  = h
    tagvals["image_height"] = w
    return (mat, tagvals)
# END tiffread