#!/usr/bin/env python
# -*- coding: utf-8 -*-
# coding=utf-8
# savecsel.py
# Almut Stribeck almut@stribeck.de
from __future__ import print_function
from matplotlib.lines import Line2D
from matplotlib.widgets import AxesWidget
import sys
import numpy as np

class VSelector( object ):
    """
    GENERAL:
    Select from a 2D window something that can be expressed by a vector.
    The vector is returned in units of image pixels counting from left
    to right and from top to bottom. The numbers are not integers, but
    floats.

    If applied to figures made from plotted curves, the positions are
    returned in data coordinates of the curve itself. Then, in contrary
    to the image case, the mathematical convention holds and the origin
    is the origin of the coordinate system with coordinates increasing
    from left to right and from bottom to top.

    The vector is input by pressing the left mouse key at the start point
    and releasing it at the end point. Switching to a different release
    key signals user abort.

    ADVANTAGES:
      * No re-organization of the user input as with the widget selectors
        of original matplotlib after version 1.3.1
      * Works with several versions of matplotlib (hopefully with all)
      * With the Ctrl-key pressed, the vector or input region is moved
        as a whole.

    EXAMPLE USAGE:

      from __future__ import print_function
      import matplotlib.pyplot as plt
      ax = plt.subplot(111) # from a 1x1 grid in the window choose subplot 1
      ax.plot(x,y) # x and y should represent a curve

      def onselect(eclick, erelease, roi):
          #eclick and erelease are matplotlib events at press and release
          print('startposition : (%f,%f)'%(eclick.xdata, eclick.ydata))
          print('endposition   : (%f,%f)'%(erelease.xdata, erelease.ydata))
          print('used button   : %s'%eclick.button)

      span = VSelector(ax, onselect,flavor='line',useblit=True)
      plt.show()

    VARIANTS:
    flavor  ='none':   The user does not see the vector on screen
            ='line':   The vector itself is shown during drawing.
            ='box':    The vector is interpreted as the diagonal of a box
                         and this box is shown
            ='circle': The vector is interpreted as the radius vector of a
                         circle and the respective circle is shown
            ='open_bar':
            ='ring':

    HISTORY: This VSelector is a slimmed and extended version of the
    RectangleSelector from matplotlib 1.3.1 which still kept the positions
    of button press and button release in the order in which they are
    entered by the user. The function has been tested under matplotlib 1.3.1
    which came with my python 2.7 and the newest version 2.1.0 which came
    with python 3.6.3 (ipy:
      In [1]: import matplotlib
      In [2]: matplotlib.__version__
      Out[1]: 2.1.0

    I did not define it as a subclass of AxesWidget, because then I would have
    to check against different versions. Moreover, I did not treat rectangles
    as something different from a line (patch vs. Line2D). On the abstract
    level there is no difference between lines, boxes and circles.
    almut@stribeck.de January 2018
    """
    def __init__(self, ax, onselect, flavor='line', useblit=False,
                 lineprops=None,feed=None):

        """
        Create a vector selector in ax.  When a selection is made, clear
        the span and call onselect with

          onselect(pos_1, pos_2, roi)

        and clear the drawn line. There pos_i are arrays of length 2
        containing the x- and y-coordinate. In general, the roi is a
        closed polygon (i.e. the first and the last point are identical).
        It is stored as two 1D-arrays with: roi[0] the x-values and
        roi[1] the y-values. For rings we will (future) have a second
        polygon stored in roi[2] and roi[3].

        The line is drawn with lineprops; default
          lineprops = dict(color='green', linestyle='-',
                           linewidth = 2, alpha=1.0)
        """
        self.ax = ax
        self.active = True
        self.visible = True
        self.set_flavor( flavor, feed )
        self._roi = None # Region of interest described by the last call
        self._roiis = '' # Flavor of the last call
        self._ringxc, self._ringyc = None, None # For ring input

        self.canvas = ax.figure.canvas
        self.cid_mmov = AxesWidget(self.ax).connect_event('motion_notify_event', self.onmove)
        self.cid_mpre = AxesWidget(self.ax).connect_event('button_press_event', self.press)
        self.cid_mrel = AxesWidget(self.ax).connect_event('button_release_event', self.release)
        self.cid_draw = AxesWidget(self.ax).connect_event('draw_event', self.update_background)
        self.cid_kpre = AxesWidget(self.ax).connect_event('key_press_event', self.onkeypress)
        self.cid_krel = AxesWidget(self.ax).connect_event('key_release_event', self.onkeyrelease)

        self.to_draw = None
        self.background = None
        self.onselect = onselect # store pointer to the outer event handler
        self.useblit = useblit
        self.ctrldown = False # Is the "ctrl"-key down?
        self.altdown = False # Is the "alt"-key down?
        # will save the data (position at mouseclick)
        self.eventpress = None
        # will save the data (pos. at mouserelease)
        self.eventrelease = None

        if lineprops is None:
            lineprops = dict(color='green', linestyle='-',
                             linewidth = 2, alpha=1.0)
        self.lineprops = lineprops
        self.xact = 0.0  # actual mouse position
        self.yact = 0.0
        # Generate the unit circle represented by a polygon
        # for fast drawing of circles
        self.sides = 72
        self._ugx, self._ugy = self._unigon( self.sides )

        # Cases mostly for documentation and exclusion of bad flavors
        if self._flavor == 'open_bar':
            # A bar is an inclined box. Its spine is the line of a previous
            # call and must be fed via self.set_flavor( 'open_bar', feed=self._roi )
            raise ValueError('VSelector must not be initiated with "open_bar"')
            # A bar is  a polygon of 5 vertices (last for closing the polygon)
        elif self._flavor == 'box':
            # Box will be represented by a polygon of 5 nodes
            # with sides parallel to the edges of the image, e.g.:
            # xl = [0, 0, 0, 0, 0]
            pass
        elif self._flavor == 'line':
            # NOTICE drawtype 'none' is 'line' with self.visibility=False
            # Line by start point and end point
            # xl, yl = [0,0], [0,0]
            pass
        elif self._flavor == 'circle':
            # circle by center and radius
            pass
        elif self._flavor == 'ring':
            # A ring (two circles) with its center in the center of the image
            # Compute the coordinates of the image center
            raise ValueError('VSelector must not be initiated with "ring"')
        else:
            raise ValueError('flavor "%s" unknown'%flavor)

        # Food for Line2D
        self._x2D, self._y2D = [0,0], [0,0]

        self.to_draw = Line2D(self._x2D, self._y2D, visible=False,**self.lineprops)
        self.ax.add_line(self.to_draw)
    #END __init__ of VSelector

    def _unigon( self, sides ):
        """
        Representation of the unit circle by a regular polygon
        """
        xl = np.zeros( sides+1 )
        yl = np.zeros( sides+1 )
        for node in range( sides+1 ):
            phi = 2.0*np.pi*node/sides
            xl[node] = np.cos(phi)
            yl[node] = np.sin(phi)
        return xl, yl
    #END _unigon() of VSelector

    def set_flavor(self, flavor, feed=None):
        """
        Specify the flavor of the selector
        """
        if flavor in ('none','line','box','open_bar','circle', 'ring'):
            self.visible = True
            self._flavor = flavor
            if flavor == 'none':
                self._flavor = 'line'
                self.visible = False
        else:
            raise ValueError('VSelector: unknown flavor "%s"'%flavor)

        if self._flavor == 'ring':
            # A ring (two circles) with its center in the center of the image
            # Compute the coordinates of the image center. For 512 x 512, e.g.:
            xmin, xmax  = self.ax.get_xlim() # -0.5, 511.5
            # print( 'xmin, xmax: %e, %e'%(xmin,xmax) )
            self._ringxc = (xmax + xmin)/2    # 511.5, -0.5
            ymin, ymax  = self.ax.get_ylim()
            # print( ymin, ymax )
            self._ringyc = (ymax + ymin)/2
            # return self._ringxc, self._ringyc #TEST
            return
        if flavor == 'open_bar':
            if feed is None:
                raise ValueError('VSelector: "open_bar": "feed" missing!')
            self._spx = feed[0]
            self._spy = feed[1]
            self._spdx = np.float(1.0)*self._spx[1] - self._spx[0]
            self._spdy = np.float(1.0)*self._spy[1] - self._spy[0]
            # Test for degenerate cases (spine too flat, spine too steep)
            steepmax = 100.0
            dc = (abs(self._spdx) > steepmax*abs(self._spdy)) or (abs(self._spdy) > steepmax*abs(self._spdx))
            if dc:
                print('SORRY! Degenerate "bar". Use "box", instead! (already selected)')
                self._flavor = 'box'
                return
            # Prepare the opening of the bar
            # Compute principal form of the (linear) spine equation
            #     y = m x + b
            self._spm = self._spdy /self._spdx
            self._spb = self._spy[0] - self._spm*self._spx[0]
            # slope at right angles
            self._spmr = - 1.0/self._spm
            # Bar width is still zero
            self._spw = 0.0
            # This is all we need to construct and show the initial polygon

    #END set_flavor() of VSelector

    def set_active(self, active):
        """
        Use this to activate / deactivate the VSelector
        from your program with an boolean parameter *active*.
        """
        self.active = active
        if active:
            # Reset food arrays/lists for Line2D
            #   otherwise the last input will flash up first
            self._x2D, self._y2D = [0,0], [0,0]
    #END set_active() of VSelector

    def get_active(self):
        """ Get status of active mode (boolean variable)"""
        return self.active
    #END get_active() of VSelector

    def update_background(self, event):
        'force an update of the background'
        if self.useblit:
            self.background = self.canvas.copy_from_bbox(self.ax.bbox)
    #END update_background() of VSelector

    def ignore(self, event):
        """return *True* if *event* should be ignored"""
        if not self.active:
            return True
    #END ignore() of VSelector

    def press(self, event):
        'on button press event'
        # Is the correct button pressed within the correct axes?
        if self.ignore(event): return

        # make the drawn line visible and get the click-coordinates,
        # button, ...
        self.to_draw.set_visible(self.visible)

        if event.inaxes != self.ax:
            # Event was not in axes. So xdata and ydata are None. We must
            # generate xdata, ydata by transformation of x, y
            (xdata, ydata) = self.ax.transData.inverted().transform_point(
                (event.x, event.y))
            # Now xdata and ydata may be outside the scope of the actual view.
            # Limiting is unnecessary.

            event.xdata = xdata
            event.ydata = ydata

        self.eventpress = event
        return False
    #END press() of VSelector

    def release(self, event):
        'on button release event'
        if self.eventpress is None or self.ignore(event): return
        # make the "line" invisible again
        self.to_draw.set_visible(False)
        self.canvas.draw()

        if self.altdown:
            # User wishes to exit from action
            self.altdown = False
            self.eventpress = None
            self.eventrelease = event # Need it to access window, if interactive
            print('Mouse input: User abort by holding the "Alt"-key during button release')
            self.onselect(self.eventpress, self.eventrelease, self._roi)
            return False

        if event.inaxes != self.ax:
            # Event was not in axes. So xdata and ydata are None. We must
            # transform from x, y to xdata, ydata
            (xdata, ydata) = self.ax.transData.inverted().transform_point(
                (event.x, event.y))
            # Now xdata and ydata may be outside the scope of the actual view.
            # Limiting is unnecessary
            event.xdata = xdata
            event.ydata = ydata

        # Transfer inner event to outer one
        self.eventrelease = event

        # Compute the selected region of interest (roi)
        x1, x2 = self.eventpress.xdata, self.eventrelease.xdata
        y1, y2 = self.eventpress.ydata, self.eventrelease.ydata
        if self._flavor == 'box':
            # The box as a polygon; the last point closes it.
            x = [x1,x2,x2,x1,x1]
            y = [y1,y1,y2,y2,y1]
        elif self._flavor == 'circle':
            #NOTICE here we work with numpy arrays ("vector operations")
            x = self._x2D # closed polygon
            y = self._y2D
        elif self._flavor == 'open_bar':
            x = self._x2D
            y = self._y2D
        elif self._flavor == 'ring':
            # Extract the two rings from the Line2D-chains
            rl = len(self._x2D)/2 # ring length in segments
            xp, xr = self._x2D[:rl], self._x2D[rl+1:]
            yp, yr = self._y2D[:rl], self._y2D[rl+1:]
            xc, yc = self._ringxc, self._ringyc # Image center
            rp = np.sqrt((x1-xc)**2 + (y1-yc)**2) # radius of press-circle
            rr = np.sqrt((x2-xc)**2 + (y2-yc)**2)
            # Place the smaller circle in the first roi
            # and the bigger in the second
            if rp < rr:
                self._roi = [xp, yp, xr, yr]
            else:
                self._roi = [xr, yr, xp, yp]
                # NOTICE Invert the vector, respectively
                self.eventpress.xdata = x2
                self.eventpress.ydata = y2
                self.eventrelease.xdata = x1
                self.eventrelease.ydata = y1
                # Now "press" is ALWAYS related to the smaller circle.
        else:
            # pass the line as a "two-angle" closed polygon
            x = [x1, x2, x1]
            y = [y1, y2, y1]

        # Internal bookkeeping: Store the result of this call
        # NOTICE Frome here it can be fetched before a follow-up
        #        call (and, e.g., fed to 'feed' for 'open_bar')
        self._roiis = self._flavor
        if self._flavor != 'ring':
            self._roi = [ x, y ]

        # call the user's method
        self.onselect(self.eventpress, self.eventrelease, self._roi)
        # reset the events
        self.eventpress = None
        self.eventrelease = None
        return False
    #END release() of VSelector

    def update(self):
        'draw using newfangled blit or oldfangled draw depending on useblit'
        if self.useblit:
            if self.background is not None:
                self.canvas.restore_region(self.background)
            self.ax.draw_artist(self.to_draw)
            self.canvas.blit(self.ax.bbox)
        else:
            self.canvas.draw_idle()
        return False
    #END update() of VSelector

    def onmove(self, event):
        'on motion notify event update the vector and draw its representation'
        if (self.eventpress is None) or (self.ignore(event)):
            return
        if (event.xdata is None) or (event.ydata is None):
            incx, incy = 0.0, 0.0
        else:
            incx = event.xdata - self.xact
            incy = event.ydata - self.yact
            # Always track the end point (where the pointer is)
            self.xact, self.yact = event.xdata, event.ydata  # actual position (with button still pressed)
        if self.ctrldown:
            if self._flavor != 'open_bar':
                # Also track the starting point of the line
                self.eventpress.xdata = self.eventpress.xdata + incx
                self.eventpress.ydata = self.eventpress.ydata + incy
                # NOTICE The starting point may be moved out of the Axes

        # Discriminate among the draw modes
        if self._flavor == 'box':
            x1, x2 = self.eventpress.xdata, self.xact  # start and actual x
            y1, y2 = self.eventpress.ydata, self.yact  # start and actual y
            #NOTICE here we have simple lists
            self._x2D = [x1,x2,x2,x1,x1]
            self._y2D = [y1,y1,y2,y2,y1]
        elif self._flavor == 'line':
            #NOTICE here we have simple lists
            self._x2D = [self.eventpress.xdata, self.xact]
            self._y2D = [self.eventpress.ydata, self.yact]
        elif self._flavor == 'circle':
            xc, yc = self.eventpress.xdata, self.eventpress.ydata
            radius = np.sqrt( (self.xact - xc)**2 + (self.yact - yc)**2 )
            #NOTICE here we work with numpy arrays ("vector operations")
            self._x2D = self._ugx*radius + xc
            self._y2D = self._ugy*radius + yc
        elif self._flavor == 'ring':
            xc, yc = self._ringxc, self._ringyc # image center
            xp, yp = self.eventpress.xdata, self.eventpress.ydata
            rp = np.sqrt((xp - xc)**2 + (yp - yc)**2)
            xsp = self._ugx*rp + xc # x-segments of press circle
            ysp = self._ugy*rp + yc # y-segments of press circle
            xa, ya = self.xact, self.yact # actual vector end
            ra = np.sqrt((xa - xc)**2 + (ya - yc)**2)
            xsa = self._ugx*ra + xc # x-segments of v_a circle
            ysa = self._ugy*ra + yc # y-segments of v_a circle
            self._x2D = np.hstack( (xsp, xsa) ) # chain together for draw!
            self._y2D = np.hstack( (ysp, ysa) )
        elif self._flavor == 'open_bar':
            x1, x2 = self.eventpress.xdata, self.xact  # start and actual x
            y1, y2 = self.eventpress.ydata, self.yact  # start and actual y
            # Only quarter of the mouse travel for higher sensitivity
            w = 0.25*np.sqrt((x2-x1)**2 + (y2-y1)**2)
            # Compute vertices of the bar
            dx = np.sqrt(w**2/(1.0+self._spmr**2))
            dy = self._spmr*dx
            self._x2D = [self._spx[0]+dx,self._spx[0]-dx,self._spx[1]-dx,self._spx[1]+dx,self._spx[0]+dx]
            self._y2D = [self._spy[0]+dy,self._spy[0]-dy,self._spy[1]-dy,self._spy[1]+dy,self._spy[0]+dy]

        self.to_draw.set_data(self._x2D,self._y2D)
        self.update()
        return False
    #END onmove() of VSelector

    def onkeypress(self, event):
        'Handle keyboard events.'
        sys.stdout.flush()
        if event.key == None:
            #Shift-Lock pressed
            return
        # Drawing the mouse while "control" is down moves
        # both points of the vector
        if event.key == 'control':
            self.ctrldown = True
        # Mouse button release while "alt" is down signals exit
        if event.key == 'alt':
            self.altdown = True
        # print('pressed: ',event.key)
    #END onkeypress() of VSelector

    def onkeyrelease(self, event):
        'Handle keyboard events.'
        sys.stdout.flush()
        if event.key == None:
            #Shift-Lock pressed
            return
        if event.key == 'control':
            self.ctrldown = False
        if event.key == 'alt':
            self.altdown = False
        # print('released: ',event.key)
    #END onkeyrelease() of VSelector
