#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards

'''setnorm.py: Normalizes a set of curves which describe a complete experimental
run with respect to sample thickness. Simultaneously the curves can be
converted to absolute units of scattering.
SYNTAX:  @@setnorm,<frameset>[''],<thick0>['1.0'],<absfact>[1.0]
        <frameset>: A file description (with wildcards) which specifies a set
                    of data files (extension: ".dat")
        <thick0>:  The initial thickness of the sample (in the same thickness
                   units which has been used when the <absfact> was
                   computed, if an <absfact> != 1.0 is specified)
        <absfact>: The intensities will be multiplied by this factor for a
                   conversion to absolute intensity. The factor can, e.g.,
                   be computed by evaluating the scattering of a calibration
                   sample and relating its known Q_a in absolute units to
                   the found experimental scattering power Q_e:
                   <absfact> = Q_a/Q_e. Before the Q_e is computed, the
                   scattering curve is (1) absorption corrected and (2)
                   divided by the known thickness of the calibration sample.
RESULT: a new set with the letter 'n' prepended to all file names. Such sets
        are overwritten in repeated runs of the script.
EXAMPLE: @@setnorm,mbsy02_*,2.03,79.6
         takes the set defined by the files mbsy02_*.dat, associates the first
         frame to a thickness of 2.03 (here: mm). For all the other frames it
         computes the actual thickness from the attribute "mut" and carries out
         the thickness normalization. Because an <absnorm> factor is given,
         the script multiplies all intensities by 78.6 and thus turns them
         into absolute units, here: electrons/nm^3 -- short: e.u./nm^3'''


## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()


def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()


def normalize():
    '''doit does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    ### Argument parsing
    argus = len( sys.argv ) - 1
    args = argexpand( 3 ) # Provide 3 arguments

    ## (1) The frame set to be worked on
    if argus < 1:
        return '1st argument (frameset) is missing.'
    frameset = args[0]

    # If the user has given no extension, then append '.dat'
    test = frameset.split('.')
    if len(test) < 2:
        frameset = frameset+'.dat'

    # alphabetically sorted list of the specified files
    allfiles = sorted(glob.glob(frameset))
    if len(allfiles) < 1:
        return 'No files matching "%s"'%frameset

    ## (2) Sample thickness
    thicks = args[1]
    thickcorr = thicks != ''
    if thickcorr:
        try:
            thick0 = float(thicks)
        except:
            return 'Unable to convert thickness at frame 0 "%s" to number.'%thicks
        if thick0 <= 0.0:
            return 'Thickness at frame 0 "%s" is not positive'%thicks

    ## (3) absfactor
    abss = args[2]
    absnorm = abss != ''
    if absnorm:
        try:
            absfact = float(abss)
        except:
            return 'Unable to convert abs-factor "%s" to number.'%abss
        if absfact <= 0.0:
            return 'abs-factor "%s" is not positive'%abss

    ### Do the work
    if (not thickcorr) and (not absnorm):
        self.poutput( 'Nothing to do' )
        return

    # Get "µt" from the initial frame
    opera = self.ops[0]
    opera.readascii( allfiles[0] )
    key = 'mut'
    if key in opera.env:
        t0 = opera.env[key]
    else:
        return 'Thickness correction impossible. Key "%s" undefined in attribute dict.'%key

    # Define the standard LaTeX labels for the ordinate (i.e. y-axis)
    lords = '$I$' # source: intensity without thickness correction
    lord  = '$I/V$' # Thickness correction applied
    lordu = 'e.u. nm$^{{-3}}$' # Data in absolute units
    for i, thefile in enumerate(allfiles):
        opera.readascii( thefile )
        span = opera.n
        if thickcorr:
            if opera.env['axis2'] == lord:
                return 'No thickness correction, because curve is already "%s"'%lord
            else:
                if opera.env['axis2'] != lords:
                    return 'No thickness correction, because this is no "%s"-curve'%lords
                else:
                    ti = opera.env[key]
                    tinv = t0 / (ti * thick0 ) # inverse actual thickness
                    opera.y[:span] = opera.y[:span] * tinv
                    opera.env['axis2'] = lord # Normalization to volume done
        if absnorm:
            if opera.env['units2'] == lordu:
                return 'Curve is already in units "%s"'%lordu
            else:
                # We do not check, if he curve is corrected for thickness.
                opera.y[:span] = opera.y[:span] * absnorm
                opera.env['units2'] = lordu # indicate: is in absolute units
        outfile = 'n'+thefile
        self.poutput('Writing normalized curve to "%s"'%outfile)
        opera.writeascii( outfile, True ) # allow overwriting
    return
#END normalize()

# END script procedures

# Main:
err = normalize()
if err:
    self.perror( err, traceback_war=Fals )
    # No signaling of error to the caller
banner()



